#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
from typing import Union

from hsml import client, model, tag
from hsml.core import explicit_provenance


class ModelApi:
    def __init__(self):
        pass

    def put(self, model_instance, query_params):
        """Save model metadata to the model registry.

        :param model_instance: metadata object of model to be saved
        :type model_instance: Model
        :return: updated metadata object of the model
        :rtype: Model
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "modelregistries",
            str(model_instance.model_registry_id),
            "models",
            model_instance.name + "_" + str(model_instance.version),
        ]
        headers = {"content-type": "application/json"}
        return model_instance.update_from_response_json(
            _client._send_request(
                "PUT",
                path_params,
                headers=headers,
                query_params=query_params,
                data=model_instance.json(),
            )
        )

    def get(self, name, version, model_registry_id, shared_registry_project_name=None):
        """Get the metadata of a model with a certain name and version.

        :param name: name of the model
        :type name: str
        :param version: version of the model
        :type version: int
        :return: model metadata object
        :rtype: Model
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "modelregistries",
            model_registry_id,
            "models",
            name + "_" + str(version),
        ]
        query_params = {"expand": "trainingdatasets"}

        model_json = _client._send_request("GET", path_params, query_params)
        model_meta = model.Model.from_response_json(model_json)

        model_meta.shared_registry_project_name = shared_registry_project_name

        return model_meta

    def get_models(
        self,
        name,
        model_registry_id,
        shared_registry_project_name=None,
        metric=None,
        direction=None,
    ):
        """Get the metadata of models based on the name or optionally the best model given a metric and direction.

        :param name: name of the model
        :type name: str
        :param metric: Name of the metric to maximize or minimize
        :type metric: str
        :param direction: Whether to maximize or minimize the metric, allowed values are 'max' or 'min'
        :type direction: str
        :return: model metadata object
        :rtype: Model
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "modelregistries",
            model_registry_id,
            "models",
        ]
        query_params = {
            "expand": "trainingdatasets",
            "filter_by": ["name_eq:" + name],
        }

        if metric is not None and direction is not None:
            if direction.lower() == "max":
                direction = "desc"
            elif direction.lower() == "min":
                direction = "asc"

            query_params["sort_by"] = metric + ":" + direction
            query_params["limit"] = "1"

        model_json = _client._send_request("GET", path_params, query_params)
        models_meta = model.Model.from_response_json(model_json)

        for model_meta in models_meta:
            model_meta.shared_registry_project_name = shared_registry_project_name

        return models_meta

    def delete(self, model_instance):
        """Delete the model and metadata.

        :param model_instance: metadata object of model to delete
        :type model_instance: Model
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "modelregistries",
            str(model_instance.model_registry_id),
            "models",
            model_instance.id,
        ]
        _client._send_request("DELETE", path_params)

    def set_tag(self, model_instance, name, value: Union[str, dict]):
        """Attach a name/value tag to a model.

        A tag consists of a name/value pair. Tag names are unique identifiers.
        The value of a tag can be any valid json - primitives, arrays or json objects.

        :param model_instance: model instance to attach tag
        :type model_instance: Model
        :param name: name of the tag to be added
        :type name: str
        :param value: value of the tag to be added
        :type value: str or dict
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "modelregistries",
            str(model_instance.model_registry_id),
            "models",
            model_instance.id,
            "tags",
            name,
        ]
        headers = {"content-type": "application/json"}
        json_value = json.dumps(value)
        _client._send_request("PUT", path_params, headers=headers, data=json_value)

    def delete_tag(self, model_instance, name):
        """Delete a tag.

        Tag names are unique identifiers.

        :param model_instance: model instance to delete tag from
        :type model_instance: Model
        :param name: name of the tag to be removed
        :type name: str
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "modelregistries",
            str(model_instance.model_registry_id),
            "models",
            model_instance.id,
            "tags",
            name,
        ]
        _client._send_request("DELETE", path_params)

    def get_tags(self, model_instance, name: str = None):
        """Get the tags.

        Gets all tags if no tag name is specified.

        :param model_instance: model instance to get the tags from
        :type model_instance: Model
        :param name: tag name
        :type name: str
        :return: dict of tag name/values
        :rtype: dict
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "modelregistries",
            str(model_instance.model_registry_id),
            "models",
            model_instance.id,
            "tags",
        ]

        if name is not None:
            path_params.append(name)

        return {
            tag._name: json.loads(tag._value)
            for tag in tag.Tag.from_response_json(
                _client._send_request("GET", path_params)
            )
        }

    def get_feature_view_provenance(self, model_instance):
        """Get the parent feature view of this model, based on explicit provenance.
        These feature views can be accessible, deleted or inaccessible.
        For deleted and inaccessible feature views, only a minimal information is returned.

        # Arguments
            model_instance: Metadata object of model.

        # Returns
            `ExplicitProvenance.Links`:  the feature view used to generate this model
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "modelregistries",
            str(model_instance.model_registry_id),
            "models",
            model_instance.id,
            "provenance",
            "links",
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 2,
            "downstreamLvls": 0,
        }
        links_json = _client._send_request("GET", path_params, query_params)
        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.UPSTREAM,
            explicit_provenance.Links.Type.FEATURE_VIEW,
        )

    def get_training_dataset_provenance(self, model_instance):
        """Get the parent training dataset of this model, based on explicit provenance.
        These training datasets can be accessible, deleted or inaccessible.
        For deleted and inaccessible training dataset, only a minimal information is returned.

        # Arguments
            model_instance: Metadata object of model.

        # Returns
            `ExplicitProvenance.Links`:  the training dataset used to generate this model
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "modelregistries",
            str(model_instance.model_registry_id),
            "models",
            model_instance.id,
            "provenance",
            "links",
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 1,
            "downstreamLvls": 0,
        }
        links_json = _client._send_request("GET", path_params, query_params)
        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.UPSTREAM,
            explicit_provenance.Links.Type.TRAINING_DATASET,
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import random
from typing import List, Optional

import humps
from hsml import util


class InferenceEndpointPort:
    """Port of an inference endpoint.

    # Arguments
        name: Name of the port. It typically defines the purpose of the port (e.g., HTTP, HTTPS, STATUS-PORT, TLS)
        number: Port number.
    # Returns
        `InferenceEndpointPort`. Port of an inference endpoint.
    """

    def __init__(self, name: str, number: int, **kwargs):
        self._name = name
        self._number = number

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls.from_json(json_decamelized)

    @classmethod
    def from_json(cls, json_decamelized):
        return InferenceEndpointPort(**cls.extract_fields_from_json(json_decamelized))

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        kwargs = {}
        kwargs["name"] = util.extract_field_from_json(json_decamelized, "name")
        kwargs["number"] = util.extract_field_from_json(json_decamelized, "number")
        return kwargs

    def to_dict(self):
        return {"name": self._name, "number": self._number}

    @property
    def name(self):
        """Name of the inference endpoint port."""
        return self._name

    @property
    def number(self):
        """Port number of the inference endpoint port."""
        return self._number

    def __repr__(self):
        return f"InferenceEndpointPort(name: {self._name!r})"


class InferenceEndpoint:
    """Inference endpoint available in the current project for model inference.

    # Arguments
        type: Type of inference endpoint (e.g., NODE, KUBE_CLUSTER, LOAD_BALANCER).
        hosts: List of hosts of the inference endpoint.
        ports: List of ports of the inference endpoint.
    # Returns
        `InferenceEndpoint`. Inference endpoint.
    """

    def __init__(
        self,
        type: str,
        hosts: List[str],
        ports: Optional[List[InferenceEndpointPort]],
    ):
        self._type = type
        self._hosts = hosts
        self._ports = ports

    def get_any_host(self):
        """Get any host available"""
        return random.choice(self._hosts) if self._hosts is not None else None

    def get_port(self, name):
        """Get port by name"""
        if self._ports is not None:
            for port in self._ports:
                if port.name == name:
                    return port
        return None

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if isinstance(json_decamelized, list):
            if len(json_decamelized) == 0:
                return []
            return [cls.from_json(endpoint) for endpoint in json_decamelized]
        else:
            if "count" in json_decamelized:
                if json_decamelized["count"] == 0:
                    return []
                return [
                    cls.from_json(endpoint) for endpoint in json_decamelized["items"]
                ]
            return cls.from_json(json_decamelized)

    @classmethod
    def from_json(cls, json_decamelized):
        return InferenceEndpoint(**cls.extract_fields_from_json(json_decamelized))

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        kwargs = {}
        kwargs["type"] = util.extract_field_from_json(json_decamelized, "type")
        kwargs["hosts"] = util.extract_field_from_json(json_decamelized, "hosts")
        kwargs["ports"] = util.extract_field_from_json(
            obj=json_decamelized, fields="ports", as_instance_of=InferenceEndpointPort
        )
        return kwargs

    def to_dict(self):
        return {
            "type": self._type,
            "hosts": self._hosts,
            "ports": [port.to_dict() for port in self._ports],
        }

    @property
    def type(self):
        """Type of inference endpoint."""
        return self._type

    @property
    def hosts(self):
        """Hosts of the inference endpoint."""
        return self._hosts

    @property
    def ports(self):
        """Ports of the inference endpoint."""
        return self._ports

    def __repr__(self):
        return f"InferenceEndpoint(type: {self._type!r})"


def get_endpoint_by_type(endpoints, type) -> InferenceEndpoint:
    for endpoint in endpoints:
        if endpoint.type == type:
            return endpoint
    return None



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import copy
import os

import humps
from hsml import model
from hsml.constants import MODEL
from hsml.core import explicit_provenance


class TestModel:
    # from response json

    def test_from_response_json_empty(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_empty"]["response"]

        # Act
        m_lst = model.Model.from_response_json(json)

        # Assert
        assert isinstance(m_lst, list)
        assert len(m_lst) == 0

    def test_from_response_json_singleton(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_python"]["response"]
        json_camelized = humps.camelize(json)  # as returned by the backend

        # Act
        m = model.Model.from_response_json(copy.deepcopy(json_camelized))

        # Assert
        assert isinstance(m, list)
        assert len(m) == 1

        m = m[0]
        m_json = json["items"][0]

        self.assert_model(mocker, m, m_json, MODEL.FRAMEWORK_PYTHON)

    def test_from_response_json_list(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_list"]["response"]
        json_camelized = humps.camelize(json)  # as returned by the backend

        # Act
        m_lst = model.Model.from_response_json(copy.deepcopy(json_camelized))

        # Assert
        assert isinstance(m_lst, list)
        assert len(m_lst) == 2

        for i in range(len(m_lst)):
            m = m_lst[i]
            m_json = json["items"][i]
            self.assert_model(mocker, m, m_json, MODEL.FRAMEWORK_PYTHON)

    # constructor

    def test_constructor_base(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_base"]["response"]["items"][0]
        m_json = copy.deepcopy(json)
        id = m_json.pop("id")
        name = m_json.pop("name")

        # Act
        m = model.Model(id=id, name=name, **m_json)

        # Assert
        self.assert_model(mocker, m, json, None)

    def test_constructor_python(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_python"]["response"]["items"][0]
        m_json = copy.deepcopy(json)
        id = m_json.pop("id")
        name = m_json.pop("name")

        # Act
        m = model.Model(id=id, name=name, **m_json)

        # Assert
        self.assert_model(mocker, m, json, MODEL.FRAMEWORK_PYTHON)

    def test_constructor_sklearn(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_sklearn"]["response"]["items"][0]
        m_json = copy.deepcopy(json)
        id = m_json.pop("id")
        name = m_json.pop("name")

        # Act
        m = model.Model(id=id, name=name, **m_json)

        # Assert
        self.assert_model(mocker, m, json, MODEL.FRAMEWORK_SKLEARN)

    def test_constructor_tensorflow(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_tensorflow"]["response"]["items"][0]
        m_json = copy.deepcopy(json)
        id = m_json.pop("id")
        name = m_json.pop("name")

        # Act
        m = model.Model(id=id, name=name, **m_json)

        # Assert
        self.assert_model(mocker, m, json, MODEL.FRAMEWORK_TENSORFLOW)

    def test_constructor_torch(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_torch"]["response"]["items"][0]
        m_json = copy.deepcopy(json)
        id = m_json.pop("id")
        name = m_json.pop("name")

        # Act
        m = model.Model(id=id, name=name, **m_json)

        # Assert
        self.assert_model(mocker, m, json, MODEL.FRAMEWORK_TORCH)

    def test_constructor_llm(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_llm"]["response"]["items"][0]
        m_json = copy.deepcopy(json)
        id = m_json.pop("id")
        name = m_json.pop("name")

        # Act
        m = model.Model(id=id, name=name, **m_json)

        # Assert
        self.assert_model(mocker, m, json, MODEL.FRAMEWORK_LLM)

    # save

    def test_save(self, mocker, backend_fixtures):
        # Arrange
        m_json = backend_fixtures["model"]["get_python"]["response"]["items"][0]
        mock_model_engine_save = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.save"
        )
        upload_configuration = {"config": "value"}

        # Act
        m = model.Model.from_response_json(m_json)
        m.save(
            model_path="model_path",
            await_registration=1234,
            keep_original_files=True,
            upload_configuration=upload_configuration,
        )

        # Assert
        mock_model_engine_save.assert_called_once_with(
            model_instance=m,
            model_path="model_path",
            await_registration=1234,
            keep_original_files=True,
            upload_configuration=upload_configuration,
        )

    # deploy

    def test_deploy(self, mocker, backend_fixtures):
        # Arrange
        m_json = backend_fixtures["model"]["get_python"]["response"]["items"][0]
        p_json = backend_fixtures["predictor"]["get_deployments_singleton"]["response"][
            "items"
        ][0]
        mock_predictor = mocker.Mock()
        mock_predictor_for_model = mocker.patch(
            "hsml.predictor.Predictor.for_model", return_value=mock_predictor
        )
        # params
        resources = copy.deepcopy(p_json["predictor_resources"])
        inference_logger = {
            "mode": p_json["inference_logging"],
            "kafka_topic": copy.deepcopy(p_json["kafka_topic_dto"]),
        }
        inference_batcher = copy.deepcopy(p_json["batching_configuration"])
        transformer = {
            "script_file": p_json["transformer"],
            "resources": copy.deepcopy(p_json["transformer_resources"]),
        }

        # Act
        m = model.Model.from_response_json(m_json)
        m.deploy(
            name=p_json["name"],
            description=p_json["description"],
            artifact_version=p_json["artifact_version"],
            serving_tool=p_json["serving_tool"],
            script_file=p_json["predictor"],
            resources=resources,
            inference_logger=inference_logger,
            inference_batcher=inference_batcher,
            transformer=transformer,
            api_protocol=p_json["api_protocol"],
            environment=p_json["environment_dto"]["name"],
        )

        # Assert
        mock_predictor_for_model.assert_called_once_with(
            m,
            name=p_json["name"],
            description=p_json["description"],
            artifact_version=p_json["artifact_version"],
            serving_tool=p_json["serving_tool"],
            script_file=p_json["predictor"],
            resources=resources,
            inference_logger=inference_logger,
            inference_batcher=inference_batcher,
            transformer=transformer,
            api_protocol=p_json["api_protocol"],
            environment=p_json["environment_dto"]["name"],
        )
        mock_predictor.deploy.assert_called_once()

    # delete

    def test_delete(self, mocker, backend_fixtures):
        # Arrange
        m_json = backend_fixtures["model"]["get_python"]["response"]["items"][0]
        mock_model_engine_delete = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.delete"
        )

        # Act
        m = model.Model.from_response_json(m_json)
        m.delete()

        # Assert
        mock_model_engine_delete.assert_called_once_with(model_instance=m)

    # download

    def test_download(self, mocker, backend_fixtures):
        # Arrange
        m_json = backend_fixtures["model"]["get_python"]["response"]["items"][0]
        mock_model_engine_download = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.download"
        )

        # Act
        m = model.Model.from_response_json(m_json)
        m.download()

        # Assert
        mock_model_engine_download.assert_called_once_with(
            model_instance=m, local_path=None
        )

    # tags

    def test_get_tag(self, mocker, backend_fixtures):
        # Arrange
        m_json = backend_fixtures["model"]["get_python"]["response"]["items"][0]
        mock_model_engine_get_tag = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.get_tag"
        )

        # Act
        m = model.Model.from_response_json(m_json)
        m.get_tag("tag_name")

        # Assert
        mock_model_engine_get_tag.assert_called_once_with(
            model_instance=m, name="tag_name"
        )

    def test_get_tags(self, mocker, backend_fixtures):
        # Arrange
        m_json = backend_fixtures["model"]["get_python"]["response"]["items"][0]
        mock_model_engine_get_tags = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.get_tags"
        )

        # Act
        m = model.Model.from_response_json(m_json)
        m.get_tags()

        # Assert
        mock_model_engine_get_tags.assert_called_once_with(model_instance=m)

    def test_set_tag(self, mocker, backend_fixtures):
        # Arrange
        m_json = backend_fixtures["model"]["get_python"]["response"]["items"][0]
        mock_model_engine_set_tag = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.set_tag"
        )

        # Act
        m = model.Model.from_response_json(m_json)
        m.set_tag("tag_name", "tag_value")

        # Assert
        mock_model_engine_set_tag.assert_called_once_with(
            model_instance=m, name="tag_name", value="tag_value"
        )

    def test_delete_tag(self, mocker, backend_fixtures):
        # Arrange
        m_json = backend_fixtures["model"]["get_python"]["response"]["items"][0]
        mock_model_engine_delete_tag = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.delete_tag"
        )

        # Act
        m = model.Model.from_response_json(m_json)
        m.delete_tag("tag_name")

        # Assert
        mock_model_engine_delete_tag.assert_called_once_with(
            model_instance=m, name="tag_name"
        )

    # get url

    def test_get_url(self, mocker, backend_fixtures):
        # Arrange
        m_json = backend_fixtures["model"]["get_python"]["response"]["items"][0]

        class ClientMock:
            _project_id = 1

        mock_client_get_instance = mocker.patch(
            "hsml.client.get_instance", return_value=ClientMock()
        )
        mock_util_get_hostname_replaced_url = mocker.patch(
            "hsml.util.get_hostname_replaced_url", return_value="full_path"
        )
        path_arg = "/p/1/models/" + m_json["name"] + "/" + str(m_json["version"])

        # Act
        m = model.Model.from_response_json(m_json)
        url = m.get_url()

        # Assert
        assert url == "full_path"
        mock_client_get_instance.assert_called_once()
        mock_util_get_hostname_replaced_url.assert_called_once_with(sub_path=path_arg)

    # auxiliary methods
    def assert_model(self, mocker, m, m_json, model_framework):
        assert isinstance(m, model.Model)
        assert m.id == m_json["id"]
        assert m.name == m_json["name"]
        assert m.version == m_json["version"]
        assert m.created == m_json["created"]
        assert m.creator == m_json["creator"]
        assert m.description == m_json["description"]
        assert m.project_name == m_json["project_name"]
        assert m.training_metrics == m_json["metrics"]
        assert m._user_full_name == m_json["user_full_name"]
        assert m.training_dataset == m_json["training_dataset"]
        assert m.model_registry_id == m_json["model_registry_id"]

        if model_framework is None:
            assert m.framework is None
        else:
            assert m.framework == model_framework

        mock_read_json = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.read_json",
            return_value="input_example_content",
        )
        assert m.input_example == "input_example_content"
        mock_read_json.assert_called_once_with(
            model_instance=m, resource=m_json["input_example"]
        )

        mock_read_json = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.read_json",
            return_value="model_schema_content",
        )
        assert m.model_schema == "model_schema_content"
        mock_read_json.assert_called_once_with(
            model_instance=m, resource=m_json["model_schema"]
        )

        mock_read_file = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.read_file",
            return_value="program_file_content",
        )
        assert m.program == "program_file_content"
        mock_read_file.assert_called_once_with(
            model_instance=m, resource=m_json["program"]
        )

        mock_read_file = mocker.patch(
            "hsml.engine.model_engine.ModelEngine.read_file",
            return_value="env_file_content",
        )
        assert m.environment == "env_file_content"
        mock_read_file.assert_called_once_with(
            model_instance=m, resource=m_json["environment"]
        )

    def test_get_feature_view(self, mocker):
        mock_fv = mocker.Mock()
        links = explicit_provenance.Links(accessible=[mock_fv])
        mock_fv_provenance = mocker.patch(
            "hsml.model.Model.get_feature_view_provenance", return_value=links
        )
        mock_td_provenance = mocker.patch(
            "hsml.model.Model.get_training_dataset_provenance", return_value=links
        )
        mocker.patch("os.environ", return_value={})
        m = model.Model(1, "test")
        m.get_feature_view()
        mock_fv_provenance.assert_called_once()
        mock_td_provenance.assert_called_once()
        assert not mock_fv.init_serving.called
        assert not mock_fv.init_batch_scoring.called

    def test_get_feature_view_online(self, mocker):
        mock_fv = mocker.Mock()
        links = explicit_provenance.Links(accessible=[mock_fv])
        mock_fv_provenance = mocker.patch(
            "hsml.model.Model.get_feature_view_provenance", return_value=links
        )
        mock_td_provenance = mocker.patch(
            "hsml.model.Model.get_training_dataset_provenance", return_value=links
        )
        mocker.patch("os.environ", return_value={})
        m = model.Model(1, "test")
        m.get_feature_view(online=True)
        mock_fv_provenance.assert_called_once()
        mock_td_provenance.assert_called_once()
        assert mock_fv.init_serving.called
        assert not mock_fv.init_batch_scoring.called

    def test_get_feature_view_batch(self, mocker):
        mock_fv = mocker.Mock()
        links = explicit_provenance.Links(accessible=[mock_fv])
        mock_fv_provenance = mocker.patch(
            "hsml.model.Model.get_feature_view_provenance", return_value=links
        )
        mock_td_provenance = mocker.patch(
            "hsml.model.Model.get_training_dataset_provenance", return_value=links
        )
        mocker.patch("os.environ", return_value={})
        m = model.Model(1, "test")
        m.get_feature_view(online=False)
        mock_fv_provenance.assert_called_once()
        mock_td_provenance.assert_called_once()
        assert not mock_fv.init_serving.called
        assert mock_fv.init_batch_scoring.called

    def test_get_feature_view_deployment(self, mocker):
        mock_fv = mocker.Mock()
        links = explicit_provenance.Links(accessible=[mock_fv])
        mock_fv_provenance = mocker.patch(
            "hsml.model.Model.get_feature_view_provenance", return_value=links
        )
        mock_td_provenance = mocker.patch(
            "hsml.model.Model.get_training_dataset_provenance", return_value=links
        )
        mocker.patch.dict(os.environ, {"DEPLOYMENT_NAME": "test"})
        m = model.Model(1, "test")
        m.get_feature_view()
        mock_fv_provenance.assert_called_once()
        mock_td_provenance.assert_called_once()
        assert mock_fv.init_serving.called
        assert not mock_fv.init_batch_scoring.called



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import copy

import pytest
from hsml import resources
from hsml.constants import RESOURCES
from mock import call


SERVING_RESOURCE_LIMITS = {"cores": 2, "memory": 516, "gpus": 2}


class TestResources:
    # Resources

    def test_from_response_json_cpus(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["resources"]["get_only_cores"]["response"]

        # Act
        r = resources.Resources.from_response_json(json)

        # Assert
        assert r.cores == json["cores"]
        assert r.memory is None
        assert r.gpus is None

    def test_from_response_json_memory(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["resources"]["get_only_memory"]["response"]

        # Act
        r = resources.Resources.from_response_json(json)

        # Assert
        assert r.cores is None
        assert r.memory is json["memory"]
        assert r.gpus is None

    def test_from_response_json_gpus(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["resources"]["get_only_gpus"]["response"]

        # Act
        r = resources.Resources.from_response_json(json)

        # Assert
        assert r.cores is None
        assert r.memory is None
        assert r.gpus == json["gpus"]

    def test_from_response_json_cores_and_memory(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["resources"]["get_cores_and_memory"]["response"]

        # Act
        r = resources.Resources.from_response_json(json)

        # Assert
        assert r.cores == json["cores"]
        assert r.memory == json["memory"]
        assert r.gpus is None

    def test_from_response_json_cores_and_gpus(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["resources"]["get_cores_and_gpus"]["response"]

        # Act
        r = resources.Resources.from_response_json(json)

        # Assert
        assert r.cores == json["cores"]
        assert r.memory is None
        assert r.gpus == json["gpus"]

    def test_from_response_json_memory_and_gpus(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["resources"]["get_memory_and_gpus"]["response"]

        # Act
        r = resources.Resources.from_response_json(json)

        # Assert
        assert r.cores is None
        assert r.memory == json["memory"]
        assert r.gpus == json["gpus"]

    def test_from_response_json_cores_memory_and_gpus(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["resources"]["get_cores_memory_and_gpus"]["response"]

        # Act
        r = resources.Resources.from_response_json(json)

        # Assert
        assert r.cores == json["cores"]
        assert r.memory == json["memory"]
        assert r.gpus == json["gpus"]

    # ComponentResources

    # - from response json

    def test_from_response_json_component_resources(self, mocker):
        # Arrange
        res = {"something": "here"}
        json_decamelized = {"key": "value"}
        mock_humps_decamelize = mocker.patch(
            "humps.decamelize", return_value=json_decamelized
        )
        mock_from_json = mocker.patch(
            "hsml.resources.ComponentResources.from_json",
            return_value="from_json_result",
        )

        # Act
        result = resources.ComponentResources.from_response_json(res)

        # Assert
        assert result == "from_json_result"
        mock_humps_decamelize.assert_called_once_with(res)
        mock_from_json.assert_called_once_with(json_decamelized)

    # - constructor

    def test_constructor_component_resources_default(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["resources"][
            "get_component_resources_num_instances_requests_and_limits"
        ]["response"]
        mock_default_resource_limits = mocker.patch(
            "hsml.resources.ComponentResources._get_default_resource_limits",
            return_value=(0, 1, 2),
        )
        mock_fill_missing_resources = mocker.patch(
            "hsml.resources.ComponentResources._fill_missing_resources"
        )
        mock_validate_resources = mocker.patch(
            "hsml.resources.ComponentResources._validate_resources"
        )
        mock_resources_init = mocker.patch(
            "hsml.resources.Resources.__init__", return_value=None
        )

        # Act
        pr = resources.PredictorResources(num_instances=json["num_instances"])

        # Assert
        assert pr.num_instances == json["num_instances"]
        assert mock_default_resource_limits.call_count == 2
        assert mock_fill_missing_resources.call_count == 2
        assert (
            mock_fill_missing_resources.call_args_list[0][0][1] == RESOURCES.MIN_CORES
        )
        assert (
            mock_fill_missing_resources.call_args_list[0][0][2] == RESOURCES.MIN_MEMORY
        )
        assert mock_fill_missing_resources.call_args_list[0][0][3] == RESOURCES.MIN_GPUS
        assert mock_fill_missing_resources.call_args_list[1][0][1] == 0
        assert mock_fill_missing_resources.call_args_list[1][0][2] == 1
        assert mock_fill_missing_resources.call_args_list[1][0][3] == 2
        mock_validate_resources.assert_called_once_with(pr._requests, pr._limits)
        expected_calls = [
            call(RESOURCES.MIN_CORES, RESOURCES.MIN_MEMORY, RESOURCES.MIN_GPUS),
            call(0, 1, 2),
        ]
        mock_resources_init.assert_has_calls(expected_calls)

    def test_constructor_component_resources(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["resources"][
            "get_component_resources_num_instances_requests_and_limits"
        ]["response"]
        mock_util_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json",
            side_effect=[json["requests"], json["limits"]],
        )
        mock_default_resource_limits = mocker.patch(
            "hsml.resources.ComponentResources._get_default_resource_limits",
            return_value=(0, 1, 2),
        )
        mock_fill_missing_resources = mocker.patch(
            "hsml.resources.ComponentResources._fill_missing_resources"
        )
        mock_validate_resources = mocker.patch(
            "hsml.resources.ComponentResources._validate_resources"
        )

        # Act
        pr = resources.PredictorResources(
            num_instances=json["num_instances"],
            requests=json["requests"],
            limits=json["limits"],
        )

        # Assert
        assert pr.num_instances == json["num_instances"]
        assert pr.requests == json["requests"]
        assert pr.limits == json["limits"]
        mock_default_resource_limits.assert_called_once()
        assert mock_fill_missing_resources.call_count == 2
        assert (
            mock_fill_missing_resources.call_args_list[0][0][1] == RESOURCES.MIN_CORES
        )
        assert (
            mock_fill_missing_resources.call_args_list[0][0][2] == RESOURCES.MIN_MEMORY
        )
        assert mock_fill_missing_resources.call_args_list[0][0][3] == RESOURCES.MIN_GPUS
        assert mock_fill_missing_resources.call_args_list[1][0][1] == 0
        assert mock_fill_missing_resources.call_args_list[1][0][2] == 1
        assert mock_fill_missing_resources.call_args_list[1][0][3] == 2
        mock_validate_resources.assert_called_once_with(pr._requests, pr._limits)
        assert mock_util_get_obj_from_json.call_count == 2
        expected_calls = [
            call(json["requests"], resources.Resources),
            call(json["limits"], resources.Resources),
        ]
        mock_util_get_obj_from_json.assert_has_calls(expected_calls)

    # - extract fields from json

    def test_extract_fields_from_json_component_resources_with_key(
        self, backend_fixtures
    ):
        # Arrange
        json = backend_fixtures["resources"][
            "get_component_resources_requested_instances_and_predictor_resources"
        ]["response"]
        copy_json = copy.deepcopy(json)
        resources.ComponentResources.RESOURCES_CONFIG_KEY = "predictor_resources"
        resources.ComponentResources.NUM_INSTANCES_KEY = "requested_instances"

        # Act
        kwargs = resources.ComponentResources.extract_fields_from_json(copy_json)

        # Assert
        assert kwargs["num_instances"] == json["requested_instances"]
        assert isinstance(kwargs["requests"], resources.Resources)
        assert (
            kwargs["requests"].cores == json["predictor_resources"]["requests"]["cores"]
        )
        assert (
            kwargs["requests"].memory
            == json["predictor_resources"]["requests"]["memory"]
        )
        assert (
            kwargs["requests"].gpus == json["predictor_resources"]["requests"]["gpus"]
        )
        assert isinstance(kwargs["limits"], resources.Resources)
        assert kwargs["limits"].cores == json["predictor_resources"]["limits"]["cores"]
        assert (
            kwargs["limits"].memory == json["predictor_resources"]["limits"]["memory"]
        )
        assert kwargs["limits"].gpus == json["predictor_resources"]["limits"]["gpus"]

    def test_extract_fields_from_json_component_resources(
        self, mocker, backend_fixtures
    ):
        # Arrange
        json = backend_fixtures["resources"][
            "get_component_resources_requested_instances_and_predictor_resources_alternative"
        ]["response"]
        copy_json = copy.deepcopy(json)
        resources.ComponentResources.RESOURCES_CONFIG_KEY = "predictor_resources"
        resources.ComponentResources.NUM_INSTANCES_KEY = "requested_instances"

        # Act
        kwargs = resources.ComponentResources.extract_fields_from_json(copy_json)

        # Assert
        assert kwargs["num_instances"] == json["num_instances"]
        assert isinstance(kwargs["requests"], resources.Resources)
        assert kwargs["requests"].cores == json["resources"]["requests"]["cores"]
        assert kwargs["requests"].memory == json["resources"]["requests"]["memory"]
        assert kwargs["requests"].gpus == json["resources"]["requests"]["gpus"]
        assert isinstance(kwargs["limits"], resources.Resources)
        assert kwargs["limits"].cores == json["resources"]["limits"]["cores"]
        assert kwargs["limits"].memory == json["resources"]["limits"]["memory"]
        assert kwargs["limits"].gpus == json["resources"]["limits"]["gpus"]

    def test_extract_fields_from_json_component_resources_flatten(
        self, backend_fixtures
    ):
        # Arrange
        json = backend_fixtures["resources"][
            "get_component_resources_num_instances_requests_and_limits"
        ]["response"]
        copy_json = copy.deepcopy(json)
        resources.ComponentResources.RESOURCES_CONFIG_KEY = "predictor_resources"
        resources.ComponentResources.NUM_INSTANCES_KEY = "requested_instances"

        # Act
        kwargs = resources.ComponentResources.extract_fields_from_json(copy_json)

        # Assert
        assert kwargs["num_instances"] == json["num_instances"]
        assert isinstance(kwargs["requests"], resources.Resources)
        assert kwargs["requests"].cores == json["requests"]["cores"]
        assert kwargs["requests"].memory == json["requests"]["memory"]
        assert kwargs["requests"].gpus == json["requests"]["gpus"]
        assert isinstance(kwargs["limits"], resources.Resources)
        assert kwargs["limits"].cores == json["limits"]["cores"]
        assert kwargs["limits"].memory == json["limits"]["memory"]
        assert kwargs["limits"].gpus == json["limits"]["gpus"]

    # - fill missing dependencies

    def test_fill_missing_dependencies_none(self, mocker):
        # Arrange
        class MockResources:
            cores = None
            memory = None
            gpus = None

        mock_resource = MockResources()

        # Act
        resources.ComponentResources._fill_missing_resources(mock_resource, 10, 11, 12)

        # Assert
        assert mock_resource.cores == 10
        assert mock_resource.memory == 11
        assert mock_resource.gpus == 12

    def test_fill_missing_dependencies_all(self, mocker):
        # Arrange
        class MockResources:
            cores = 1
            memory = 2
            gpus = 3

        mock_resource = MockResources()

        # Act
        resources.ComponentResources._fill_missing_resources(mock_resource, 10, 11, 12)

        # Assert
        assert mock_resource.cores == 1
        assert mock_resource.memory == 2
        assert mock_resource.gpus == 3

    def test_fill_missing_dependencies_some(self, mocker):
        # Arrange
        class MockResources:
            cores = 1
            memory = None
            gpus = None

        mock_resource = MockResources()

        # Act
        resources.ComponentResources._fill_missing_resources(mock_resource, 10, 11, 12)

        # Assert
        assert mock_resource.cores == 1
        assert mock_resource.memory == 11
        assert mock_resource.gpus == 12

    # - get default resource limits

    def test_get_default_resource_limits_no_hard_limit_and_lower_than_default(
        self, mocker
    ):
        # Arrange
        no_limit_res = {"cores": -1, "memory": -1, "gpus": -1}
        mock_comp_res = mocker.MagicMock()
        mock_comp_res._requests = resources.Resources(cores=0.2, memory=516, gpus=0)
        mock_comp_res._default_resource_limits = (
            resources.ComponentResources._get_default_resource_limits
        )
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=no_limit_res,  # no upper limit
        )

        # Act
        cores, memory, gpus = mock_comp_res._default_resource_limits(mock_comp_res)

        # Assert
        assert cores == RESOURCES.MAX_CORES
        assert memory == RESOURCES.MAX_MEMORY
        assert gpus == RESOURCES.MAX_GPUS
        mock_get_serving_res_limits.assert_called_once()

    def test_get_default_resource_limits_no_hard_limit_and_higher_than_default(
        self, mocker
    ):
        # Arrange
        no_limit_res = {"cores": -1, "memory": -1, "gpus": -1}
        mock_comp_res = mocker.MagicMock()
        mock_comp_res._requests = resources.Resources(cores=4, memory=2048, gpus=2)
        mock_comp_res._default_resource_limits = (
            resources.ComponentResources._get_default_resource_limits
        )
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=no_limit_res,  # no upper limit
        )

        # Act
        cores, memory, gpus = mock_comp_res._default_resource_limits(mock_comp_res)

        # Assert
        assert cores == mock_comp_res._requests.cores
        assert memory == mock_comp_res._requests.memory
        assert gpus == mock_comp_res._requests.gpus
        mock_get_serving_res_limits.assert_called_once()

    def test_get_default_resource_limits_with_higher_hard_limit_and_lower_than_default(
        self, mocker
    ):
        # Arrange
        hard_limit_res = {"cores": 3, "memory": 3072, "gpus": 3}
        mock_comp_res = mocker.MagicMock()
        mock_comp_res._requests = resources.Resources(cores=1, memory=516, gpus=0)
        mock_comp_res._default_resource_limits = (
            resources.ComponentResources._get_default_resource_limits
        )
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        cores, memory, gpus = mock_comp_res._default_resource_limits(mock_comp_res)

        # Assert
        assert cores == RESOURCES.MAX_CORES
        assert memory == RESOURCES.MAX_MEMORY
        assert gpus == RESOURCES.MAX_GPUS
        mock_get_serving_res_limits.assert_called_once()

    def test_get_default_resource_limits_with_higher_hard_limit_and_higher_than_default(
        self, mocker
    ):
        # Arrange
        hard_limit_res = {"cores": 3, "memory": 3072, "gpus": 3}
        mock_comp_res = mocker.MagicMock()
        mock_comp_res._requests = resources.Resources(cores=3, memory=2048, gpus=1)
        mock_comp_res._default_resource_limits = (
            resources.ComponentResources._get_default_resource_limits
        )
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        cores, memory, gpus = mock_comp_res._default_resource_limits(mock_comp_res)

        # Assert
        assert cores == hard_limit_res["cores"]
        assert memory == hard_limit_res["memory"]
        assert gpus == hard_limit_res["gpus"]
        mock_get_serving_res_limits.assert_called_once()

    def test_get_default_resource_limits_with_lower_hard_limit_and_lower_than_default(
        self, mocker
    ):
        # Arrange
        RESOURCES.MAX_GPUS = 1  # override default
        hard_limit_res = {"cores": 1, "memory": 516, "gpus": 0}
        mock_comp_res = mocker.MagicMock()
        mock_comp_res._requests = resources.Resources(cores=0.5, memory=256, gpus=0.5)
        mock_comp_res._default_resource_limits = (
            resources.ComponentResources._get_default_resource_limits
        )
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        cores, memory, gpus = mock_comp_res._default_resource_limits(mock_comp_res)

        # Assert
        assert cores == hard_limit_res["cores"]
        assert memory == hard_limit_res["memory"]
        assert gpus == hard_limit_res["gpus"]
        mock_get_serving_res_limits.assert_called_once()

    def test_get_default_resource_limits_with_lower_hard_limit_and_higher_than_default(
        self, mocker
    ):
        # Arrange
        RESOURCES.MAX_GPUS = 1  # override default
        hard_limit_res = {"cores": 1, "memory": 516, "gpus": 0}
        mock_comp_res = mocker.MagicMock()
        mock_comp_res._requests = resources.Resources(cores=4, memory=4080, gpus=4)
        mock_comp_res._default_resource_limits = (
            resources.ComponentResources._get_default_resource_limits
        )
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        cores, memory, gpus = mock_comp_res._default_resource_limits(mock_comp_res)

        # Assert
        assert cores == hard_limit_res["cores"]
        assert memory == hard_limit_res["memory"]
        assert gpus == hard_limit_res["gpus"]
        mock_get_serving_res_limits.assert_called_once()

    # - validate resources

    def test_validate_resources_no_hard_limits_valid_resources(self, mocker):
        # Arrange
        no_limit_res = {"cores": -1, "memory": -1, "gpus": -1}
        requests = resources.Resources(cores=1, memory=1024, gpus=0)
        limits = resources.Resources(cores=2, memory=2048, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=no_limit_res,  # upper limit
        )

        # Act
        resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()

    def test_validate_resources_no_hard_limit_invalid_cores_request(self, mocker):
        # Arrange
        no_limit_res = {"cores": -1, "memory": -1, "gpus": -1}
        requests = resources.Resources(cores=0, memory=1024, gpus=0)
        limits = resources.Resources(cores=2, memory=2048, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=no_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert "Requested number of cores must be greater than 0 cores." in str(
            e_info.value
        )

    def test_validate_resources_no_hard_limit_invalid_memory_request(self, mocker):
        # Arrange
        no_limit_res = {"cores": -1, "memory": -1, "gpus": -1}
        requests = resources.Resources(cores=1, memory=0, gpus=0)
        limits = resources.Resources(cores=2, memory=2048, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=no_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert "Requested memory resources must be greater than 0 MB." in str(
            e_info.value
        )

    def test_validate_resources_no_hard_limit_invalid_gpus_request(self, mocker):
        # Arrange
        no_limit_res = {"cores": -1, "memory": -1, "gpus": -1}
        requests = resources.Resources(
            cores=1, memory=1024, gpus=-1
        )  # 0 gpus is accepted
        limits = resources.Resources(cores=2, memory=2048, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=no_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert (
            "Requested number of gpus must be greater than or equal to 0 gpus."
            in str(e_info.value)
        )

    def test_validate_resources_no_hard_limit_cores_request_out_of_range(self, mocker):
        # Arrange
        no_limit_res = {"cores": -1, "memory": -1, "gpus": -1}
        requests = resources.Resources(cores=2, memory=1024, gpus=0)
        limits = resources.Resources(cores=1, memory=2048, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=no_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert (
            f"Requested number of cores cannot exceed the limit of {str(limits.cores)} cores."
            in str(e_info.value)
        )

    def test_validate_resources_no_hard_limit_invalid_memory_request_out_of_range(
        self, mocker
    ):
        # Arrange
        no_limit_res = {"cores": -1, "memory": -1, "gpus": -1}
        requests = resources.Resources(cores=1, memory=2048, gpus=0)
        limits = resources.Resources(cores=2, memory=1024, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=no_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert (
            f"Requested memory resources cannot exceed the limit of {str(limits.memory)} MB."
            in str(e_info.value)
        )

    def test_validate_resources_no_hard_limit_invalid_gpus_request_out_of_range(
        self, mocker
    ):
        # Arrange
        no_limit_res = {"cores": -1, "memory": -1, "gpus": -1}
        requests = resources.Resources(cores=1, memory=1024, gpus=2)
        limits = resources.Resources(cores=2, memory=2048, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=no_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert (
            f"Requested number of gpus cannot exceed the limit of {str(limits.gpus)} gpus."
            in str(e_info.value)
        )

    def test_validate_resources_with_hard_limit_valid_resources(self, mocker):
        # Arrange
        hard_limit_res = {"cores": 3, "memory": 3072, "gpus": 3}
        requests = resources.Resources(cores=1, memory=1024, gpus=0)
        limits = resources.Resources(cores=2, memory=2048, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()

    def test_validate_resources_with_hard_limit_invalid_cores_limit(self, mocker):
        # Arrange
        hard_limit_res = {"cores": 3, "memory": 3072, "gpus": 3}
        requests = resources.Resources(cores=2, memory=1024, gpus=0)
        limits = resources.Resources(cores=0, memory=2048, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert "Limit number of cores must be greater than 0 cores." in str(
            e_info.value
        )

    def test_validate_resources_with_hard_limit_invalid_memory_limit(self, mocker):
        # Arrange
        hard_limit_res = {"cores": 3, "memory": 3072, "gpus": 3}
        requests = resources.Resources(cores=2, memory=1024, gpus=0)
        limits = resources.Resources(cores=1, memory=0, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert "Limit memory resources must be greater than 0 MB." in str(e_info.value)

    def test_validate_resources_with_hard_limit_invalid_gpus_limit(self, mocker):
        # Arrange
        hard_limit_res = {"cores": 3, "memory": 3072, "gpus": 3}
        requests = resources.Resources(cores=2, memory=1024, gpus=0)
        limits = resources.Resources(cores=1, memory=2048, gpus=-1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert "Limit number of gpus must be greater than or equal to 0 gpus." in str(
            e_info.value
        )

    def test_validate_resources_with_hard_limit_invalid_cores_request(self, mocker):
        # Arrange
        hard_limit_res = {"cores": 3, "memory": 3072, "gpus": 3}
        requests = resources.Resources(cores=2, memory=1024, gpus=0)
        limits = resources.Resources(cores=4, memory=2048, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert (
            f"Limit number of cores cannot exceed the maximum of {hard_limit_res['cores']} cores."
            in str(e_info.value)
        )

    def test_validate_resources_with_hard_limit_invalid_memory_request(self, mocker):
        # Arrange
        hard_limit_res = {"cores": 3, "memory": 3072, "gpus": 3}
        requests = resources.Resources(cores=2, memory=1024, gpus=0)
        limits = resources.Resources(cores=3, memory=4076, gpus=1)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert (
            f"Limit memory resources cannot exceed the maximum of {hard_limit_res['memory']} MB."
            in str(e_info.value)
        )

    def test_validate_resources_with_hard_limit_invalid_gpus_request(self, mocker):
        # Arrange
        hard_limit_res = {"cores": 3, "memory": 3072, "gpus": 3}
        requests = resources.Resources(cores=2, memory=1024, gpus=0)
        limits = resources.Resources(cores=3, memory=2048, gpus=4)
        mock_get_serving_res_limits = mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=hard_limit_res,  # upper limit
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            resources.ComponentResources._validate_resources(requests, limits)

        # Assert
        mock_get_serving_res_limits.assert_called_once()
        assert (
            f"Limit number of gpus cannot exceed the maximum of {hard_limit_res['gpus']} gpus."
            in str(e_info.value)
        )

    # PredictorResources

    def test_from_response_json_predictor_resources(self, mocker, backend_fixtures):
        mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=SERVING_RESOURCE_LIMITS,
        )
        json = backend_fixtures["resources"][
            "get_component_resources_num_instances_requests_and_limits"
        ]["response"]

        # Act
        r = resources.PredictorResources.from_response_json(json)

        # Assert
        assert r.num_instances == json["num_instances"]
        assert r.requests.cores == json["requests"]["cores"]
        assert r.requests.memory == json["requests"]["memory"]
        assert r.requests.gpus == json["requests"]["gpus"]
        assert r.limits.cores == json["limits"]["cores"]
        assert r.limits.memory == json["limits"]["memory"]
        assert r.limits.gpus == json["limits"]["gpus"]

    def test_from_response_json_predictor_resources_specific_keys(
        self, mocker, backend_fixtures
    ):
        mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=SERVING_RESOURCE_LIMITS,
        )
        json = backend_fixtures["resources"][
            "get_component_resources_requested_instances_and_predictor_resources"
        ]["response"]

        # Act
        r = resources.PredictorResources.from_response_json(json)

        # Assert
        assert r.num_instances == json["requested_instances"]
        assert r.requests.cores == json["predictor_resources"]["requests"]["cores"]
        assert r.requests.memory == json["predictor_resources"]["requests"]["memory"]
        assert r.requests.gpus == json["predictor_resources"]["requests"]["gpus"]
        assert r.limits.cores == json["predictor_resources"]["limits"]["cores"]
        assert r.limits.memory == json["predictor_resources"]["limits"]["memory"]
        assert r.limits.gpus == json["predictor_resources"]["limits"]["gpus"]

    # TransformerResources

    def test_from_response_json_transformer_resources(self, mocker, backend_fixtures):
        mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=SERVING_RESOURCE_LIMITS,
        )
        json = backend_fixtures["resources"][
            "get_component_resources_num_instances_requests_and_limits"
        ]["response"]

        # Act
        r = resources.TransformerResources.from_response_json(json)

        # Assert
        assert r.num_instances == json["num_instances"]
        assert r.requests.cores == json["requests"]["cores"]
        assert r.requests.memory == json["requests"]["memory"]
        assert r.requests.gpus == json["requests"]["gpus"]
        assert r.limits.cores == json["limits"]["cores"]
        assert r.limits.memory == json["limits"]["memory"]
        assert r.limits.gpus == json["limits"]["gpus"]

    def test_from_response_json_transformer_resources_specific_keys(
        self, mocker, backend_fixtures
    ):
        mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=SERVING_RESOURCE_LIMITS,
        )
        json = backend_fixtures["resources"][
            "get_component_resources_requested_instances_and_transformer_resources"
        ]["response"]

        # Act
        r = resources.TransformerResources.from_response_json(json)

        # Assert
        assert r.num_instances == json["requested_transformer_instances"]
        assert r.requests.cores == json["transformer_resources"]["requests"]["cores"]
        assert r.requests.memory == json["transformer_resources"]["requests"]["memory"]
        assert r.requests.gpus == json["transformer_resources"]["requests"]["gpus"]
        assert r.limits.cores == json["transformer_resources"]["limits"]["cores"]
        assert r.limits.memory == json["transformer_resources"]["limits"]["memory"]
        assert r.limits.gpus == json["transformer_resources"]["limits"]["gpus"]

    def test_from_response_json_transformer_resources_default_limits(
        self, mocker, backend_fixtures
    ):
        mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=SERVING_RESOURCE_LIMITS,
        )
        mocker.patch(
            "hsml.resources.ComponentResources._get_default_resource_limits",
            return_value=(
                SERVING_RESOURCE_LIMITS["cores"],
                SERVING_RESOURCE_LIMITS["memory"],
                SERVING_RESOURCE_LIMITS["gpus"],
            ),
        )
        json = backend_fixtures["resources"][
            "get_component_resources_num_instances_and_requests"
        ]["response"]

        # Act
        r = resources.TransformerResources.from_response_json(json)

        # Assert
        assert r.num_instances == json["num_instances"]
        assert r.requests.cores == json["requests"]["cores"]
        assert r.requests.memory == json["requests"]["memory"]
        assert r.requests.gpus == json["requests"]["gpus"]
        assert r.limits.cores == SERVING_RESOURCE_LIMITS["cores"]
        assert r.limits.memory == SERVING_RESOURCE_LIMITS["memory"]
        assert r.limits.gpus == SERVING_RESOURCE_LIMITS["gpus"]



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import logging
from enum import Enum
from typing import Set

import humps


_logger = logging.getLogger(__name__)


class Artifact:
    class MetaType(Enum):
        DELETED = 1
        INACCESSIBLE = 2
        FAULTY = 3
        NOT_SUPPORTED = 4

    def __init__(
        self,
        model_registry_id,
        name,
        version,
        type,
        meta_type,
        href=None,
        exception_cause=None,
        **kwargs,
    ):
        self._model_registry_id = model_registry_id
        self._name = name
        self._version = version
        self._type = type
        self._meta_type = meta_type
        self._href = href
        self._exception_cause = exception_cause

    @property
    def model_registry_id(self):
        """Id of the model registry in which the artifact is located."""
        return self._model_registry_id

    @property
    def name(self):
        """Name of the artifact."""
        return self._name

    @property
    def version(self):
        """Version of the artifact"""
        return self._version

    def __str__(self):
        return {
            "model_registry_id": self._model_registry_id,
            "name": self._name,
            "version": self._version,
        }

    def __repr__(self):
        return (
            f"Artifact({self._model_registry_id!r}, {self._name!r}, "
            f"{self._version!r}, {self._type!r}, {self._meta_type!r}, "
            f"{self._href!r}, {self._exception_cause!r})"
        )

    @staticmethod
    def from_response_json(json_dict: dict):
        link_json = humps.decamelize(json_dict)
        href = None
        exception_cause = None
        if link_json.get("exception_cause") is not None:
            meta_type = Artifact.MetaType.FAULTY
            exception_cause = link_json.get("exception_cause")
        elif bool(link_json["deleted"]):
            meta_type = Artifact.MetaType.DELETED
        elif not bool(link_json["accessible"]):
            meta_type = Artifact.MetaType.INACCESSIBLE
            href = link_json["artifact"]["href"]
        else:
            meta_type = Artifact.MetaType.NOT_SUPPORTED
            href = link_json["artifact"]["href"]
        return Artifact(
            link_json["artifact"]["project"],
            link_json["artifact"]["name"],
            link_json["artifact"]["version"],
            link_json["artifact_type"],
            meta_type,
            href=href,
            exception_cause=exception_cause,
        )


class Links:
    def __init__(self, accessible=None, deleted=None, inaccessible=None, faulty=None):
        if accessible is None:
            self._accessible = []
        else:
            self._accessible = accessible
        if deleted is None:
            self._deleted = []
        else:
            self._deleted = deleted
        if inaccessible is None:
            self._inaccessible = []
        else:
            self._inaccessible = inaccessible
        if faulty is None:
            self._faulty = []
        else:
            self._faulty = faulty

    @property
    def deleted(self):
        """List of [Artifact objects] which contains
        minimal information (name, version) about the entities
        (feature views, training datasets) they represent.
        These entities have been removed from the feature store.
        """
        return self._deleted

    @property
    def inaccessible(self):
        """List of [Artifact objects] which contains
        minimal information (name, version) about the entities
        (feature views, training datasets) they represent.
        These entities exist in the feature store, however the user
        does not have access to them anymore.
        """
        return self._inaccessible

    @property
    def accessible(self):
        """List of [FeatureView|TrainingDataset objects] objects
        which are part of the provenance graph requested. These entities
        exist in the feature store and the user has access to them.
        """
        return self._accessible

    @property
    def faulty(self):
        """List of [Artifact objects] which contains
        minimal information (name, version) about the entities
        (feature views, training datasets) they represent.
        These entities exist in the feature store, however they are corrupted.
        """
        return self._faulty

    class Direction(Enum):
        UPSTREAM = 1
        DOWNSTREAM = 2

    class Type(Enum):
        FEATURE_VIEW = 1
        TRAINING_DATASET = 2

    def __str__(self, indent=None):
        return json.dumps(self, cls=ProvenanceEncoder, indent=indent)

    def __repr__(self):
        return (
            f"Links({self._accessible!r}, {self._deleted!r}"
            f", {self._inaccessible!r}, {self._faulty!r})"
        )

    @staticmethod
    def get_one_accessible_parent(links):
        if links is None:
            _logger.info("There is no parent information")
            return
        elif links.inaccessible or links.deleted:
            _logger.info(
                "The parent is deleted or inaccessible. For more details get the full provenance from `_provenance` method"
            )
            return None
        elif links.accessible:
            if len(links.accessible) > 1:
                msg = "Backend inconsistency - provenance returned more than one parent"
                raise Exception(msg)
            parent = links.accessible[0]
            if isinstance(parent, Artifact):
                msg = "The returned object is not a valid object. For more details get the full provenance from `_provenance` method"
                raise Exception(msg)
            return parent
        else:
            _logger.info("There is no parent information")
            return None

    @staticmethod
    def __parse_feature_views(links_json: dict, artifacts: Set[str]):
        from hsfs import feature_view
        from hsfs.core import explicit_provenance as hsfs_explicit_provenance

        links = Links()
        for link_json in links_json:
            if link_json["node"]["artifact_type"] in artifacts:
                if link_json["node"].get("exception_cause") is not None:
                    links._faulty.append(
                        hsfs_explicit_provenance.Artifact.from_response_json(
                            link_json["node"]
                        )
                    )
                elif bool(link_json["node"]["accessible"]):
                    fv = feature_view.FeatureView.from_response_json(
                        link_json["node"]["artifact"]
                    )
                    links.accessible.append(fv)
                elif bool(link_json["node"]["deleted"]):
                    links.deleted.append(
                        hsfs_explicit_provenance.Artifact.from_response_json(
                            link_json["node"]
                        )
                    )
                else:
                    links.inaccessible.append(
                        hsfs_explicit_provenance.Artifact.from_response_json(
                            link_json["node"]
                        )
                    )
            else:
                new_links = Links.__parse_feature_views(
                    link_json["upstream"], artifacts
                )
                links.faulty.extend(new_links.faulty)
                links.accessible.extend(new_links.accessible)
                links.inaccessible.extend(new_links.inaccessible)
                links.deleted.extend(new_links.deleted)
        return links

    @staticmethod
    def __parse_training_datasets(links_json: dict, artifacts: Set[str]):
        from hsfs import training_dataset
        from hsfs.core import explicit_provenance as hsfs_explicit_provenance

        links = Links()
        for link_json in links_json:
            if link_json["node"]["artifact_type"] in artifacts:
                if link_json["node"].get("exception_cause") is not None:
                    links._faulty.append(
                        hsfs_explicit_provenance.Artifact.from_response_json(
                            link_json["node"]
                        )
                    )
                elif bool(link_json["node"]["accessible"]):
                    td = training_dataset.TrainingDataset.from_response_json_single(
                        link_json["node"]["artifact"]
                    )
                    links.accessible.append(td)
                elif bool(link_json["node"]["deleted"]):
                    links.deleted.append(
                        hsfs_explicit_provenance.Artifact.from_response_json(
                            link_json["node"]
                        )
                    )
                else:
                    links.inaccessible.append(
                        hsfs_explicit_provenance.Artifact.from_response_json(
                            link_json["node"]
                        )
                    )
        return links

    @staticmethod
    def from_response_json(json_dict: dict, direction: Direction, artifact: Type):
        """Parse explicit links from json response. There are three types of
        Links: UpstreamFeatureGroups, DownstreamFeatureGroups, DownstreamFeatureViews

        # Arguments
            links_json: json response from the explicit provenance endpoint
            direction: subset of links to parse - UPSTREAM/DOWNSTREAM
            type: subset of links to parse - FEATURE_VIEW/TRAINING_DATASET/MODEL

        # Returns
            A ProvenanceLink object for the selected parse type.
        """

        import importlib.util

        if not importlib.util.find_spec("hsfs"):
            raise ValueError(
                "hsfs is not installed in the environment - cannot parse feature store artifacts"
            )
        if not importlib.util.find_spec("hopsworks"):
            raise ValueError(
                "hopsworks is not installed in the environment - cannot switch from hsml connection to hsfs connection"
            )

        # make sure the hsfs connection is initialized so that the feature view/training dataset can actually be used after being returned
        import hopsworks

        if not hopsworks._connected_project:
            raise Exception(
                "hopsworks connection is not initialized - use hopsworks.login to connect if you want the ability to use provenance with connections between hsfs and hsml"
            )

        hopsworks._connected_project.get_feature_store()

        links = Links.__from_response_json_feature_store_artifacts(
            json_dict, direction, artifact
        )
        return links

    @staticmethod
    def __from_response_json_feature_store_artifacts(
        json_dict: dict, direction: Direction, artifact: Type
    ):
        links_json = humps.decamelize(json_dict)
        if direction == Links.Direction.UPSTREAM:
            if artifact == Links.Type.FEATURE_VIEW:
                return Links.__parse_feature_views(
                    links_json["upstream"],
                    {
                        "FEATURE_VIEW",
                    },
                )
            elif artifact == Links.Type.TRAINING_DATASET:
                return Links.__parse_training_datasets(
                    links_json["upstream"], {"TRAINING_DATASET"}
                )
        else:
            return Links()


class ProvenanceEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Links):
            return {
                "accessible": obj.accessible,
                "inaccessible": obj.inaccessible,
                "deleted": obj.deleted,
                "faulty": obj.faulty,
            }
        else:
            import importlib.util

            if importlib.util.find_spec("hsfs"):
                from hsfs import feature_view
                from hsfs.core import explicit_provenance as hsfs_explicit_provenance

                if isinstance(
                    obj,
                    (
                        feature_view.FeatureView,
                        hsfs_explicit_provenance.Artifact,
                    ),
                ):
                    return {
                        "feature_store_name": obj.feature_store_name,
                        "name": obj.name,
                        "version": obj.version,
                    }
            return json.JSONEncoder.default(self, obj)



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import humps
from hopsworks_common import constants, git_repo


class GitOpExecution:
    def __init__(
        self,
        id,
        submission_time,
        execution_start,
        execution_stop,
        user,
        git_command_configuration,
        state,
        config_secret,
        repository,
        command_result_message=None,
        type=None,
        href=None,
        expand=None,
        items=None,
        count=None,
        **kwargs,
    ):
        self._id = id
        self._submission_time = submission_time
        self._execution_start = execution_start
        self._execution_stop = execution_stop
        self._user = user
        self._git_command_configuration = git_command_configuration
        self._state = state
        self._command_result_message = command_result_message
        self._repository = git_repo.GitRepo.from_response_json(repository)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    @property
    def id(self):
        """Id of the execution"""
        return self._id

    @property
    def submission_time(self):
        """Timestamp when the execution was submitted"""
        return self._submission_time

    @property
    def execution_start(self):
        """Timestamp when the execution started"""
        return self._execution_start

    @property
    def execution_stop(self):
        """Timestamp when the execution stopped"""
        return self._execution_stop

    @property
    def user(self):
        """User that issued the execution"""
        return self._user

    @property
    def git_command_configuration(self):
        """Configuration for the git command"""
        return self._git_command_configuration

    @property
    def state(self):
        """State of the git execution"""
        return self._state

    @property
    def command_result_message(self):
        """Results message from the execution"""
        return self._command_result_message

    @property
    def repository(self):
        """Git repository of the execution"""
        return self._repository

    @property
    def success(self):
        """Boolean to indicate if execution ran successfully or failed

        ```
        # Returns
            `bool`. True if execution ran successfully. False if execution failed.
        """
        if self.state is not None and self.state.upper() in constants.GIT.ERROR_STATES:
            return False
        elif (
            self.state is not None
            and self.state.upper() in constants.GIT.SUCCESS_STATES
        ):
            return True
        else:
            return None



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import asyncio
import os
from datetime import date, datetime
from urllib.parse import ParseResult

import hsfs.util
import pytest
import pytz
from hsfs.client.exceptions import FeatureStoreException
from hsfs.core.constants import HAS_AIOMYSQL, HAS_SQLALCHEMY
from hsfs.embedding import EmbeddingFeature, EmbeddingIndex
from hsfs.feature import Feature
from hsml import util
from hsml.constants import MODEL
from hsml.llm.model import Model as LLMModel
from hsml.llm.predictor import Predictor as LLMPredictor
from hsml.model import Model as BaseModel
from hsml.predictor import Predictor as BasePredictor
from hsml.python.model import Model as PythonModel
from hsml.python.predictor import Predictor as PyPredictor
from hsml.sklearn.model import Model as SklearnModel
from hsml.sklearn.predictor import Predictor as SkLearnPredictor
from hsml.tensorflow.model import Model as TensorflowModel
from hsml.tensorflow.predictor import Predictor as TFPredictor
from hsml.torch.model import Model as TorchModel
from hsml.torch.predictor import Predictor as TorchPredictor
from mock import patch


if HAS_SQLALCHEMY and HAS_AIOMYSQL:
    from hsfs.core import util_sql


class TestUtil:
    # schema and types

    # - set_model_class

    def test_set_model_class_base(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_base"]["response"]["items"][0]

        # Act
        model = util.set_model_class(json)

        # Assert
        assert isinstance(model, BaseModel)
        assert model.framework is None

    def test_set_model_class_python(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_python"]["response"]["items"][0]

        # Act
        model = util.set_model_class(json)

        # Assert
        assert isinstance(model, PythonModel)
        assert model.framework == MODEL.FRAMEWORK_PYTHON

    def test_set_model_class_sklearn(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_sklearn"]["response"]["items"][0]

        # Act
        model = util.set_model_class(json)

        # Assert
        assert isinstance(model, SklearnModel)
        assert model.framework == MODEL.FRAMEWORK_SKLEARN

    def test_set_model_class_tensorflow(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_tensorflow"]["response"]["items"][0]

        # Act
        model = util.set_model_class(json)

        # Assert
        assert isinstance(model, TensorflowModel)
        assert model.framework == MODEL.FRAMEWORK_TENSORFLOW

    def test_set_model_class_torch(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_torch"]["response"]["items"][0]

        # Act
        model = util.set_model_class(json)

        # Assert
        assert isinstance(model, TorchModel)
        assert model.framework == MODEL.FRAMEWORK_TORCH

    def test_set_model_class_llm(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_llm"]["response"]["items"][0]

        # Act
        model = util.set_model_class(json)

        # Assert
        assert isinstance(model, LLMModel)
        assert model.framework == MODEL.FRAMEWORK_LLM

    def test_set_model_class_unsupported(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["model"]["get_base"]["response"]["items"][0]
        json["framework"] = "UNSUPPORTED"

        # Act
        with pytest.raises(ValueError) as e_info:
            util.set_model_class(json)

        # Assert
        assert "is not a supported framework" in str(e_info.value)

    # - input_example_to_json

    def test_input_example_to_json_from_numpy(self, mocker, input_example_numpy):
        # Arrange
        mock_handle_tensor_input = mocker.patch("hsml.util._handle_tensor_input")
        mock_handle_dataframe_input = mocker.patch("hsml.util._handle_dataframe_input")
        mock_handle_dict_input = mocker.patch("hsml.util._handle_dict_input")

        # Act
        util.input_example_to_json(input_example_numpy)

        # Assert
        mock_handle_tensor_input.assert_called_once()
        mock_handle_dict_input.assert_not_called()
        mock_handle_dataframe_input.assert_not_called()

    def test_input_example_to_json_from_dict(self, mocker, input_example_dict):
        # Arrange
        mock_handle_tensor_input = mocker.patch("hsml.util._handle_tensor_input")
        mock_handle_dataframe_input = mocker.patch("hsml.util._handle_dataframe_input")
        mock_handle_dict_input = mocker.patch("hsml.util._handle_dict_input")

        # Act
        util.input_example_to_json(input_example_dict)

        # Assert
        mock_handle_tensor_input.assert_not_called()
        mock_handle_dict_input.assert_called_once()
        mock_handle_dataframe_input.assert_not_called()

    def test_input_example_to_json_from_dataframe(
        self, mocker, input_example_dataframe_pandas_dataframe
    ):
        # Arrange
        mock_handle_tensor_input = mocker.patch("hsml.util._handle_tensor_input")
        mock_handle_dataframe_input = mocker.patch("hsml.util._handle_dataframe_input")
        mock_handle_dict_input = mocker.patch("hsml.util._handle_dict_input")

        # Act
        util.input_example_to_json(input_example_dataframe_pandas_dataframe)

        # Assert
        mock_handle_tensor_input.assert_not_called()
        mock_handle_dict_input.assert_not_called()
        mock_handle_dataframe_input.assert_called_once()  # default

    def test_input_example_to_json_unsupported(self, mocker):
        # Arrange
        mock_handle_tensor_input = mocker.patch("hsml.util._handle_tensor_input")
        mock_handle_dataframe_input = mocker.patch("hsml.util._handle_dataframe_input")
        mock_handle_dict_input = mocker.patch("hsml.util._handle_dict_input")

        # Act
        util.input_example_to_json(lambda unsupported_type: None)

        # Assert
        mock_handle_tensor_input.assert_not_called()
        mock_handle_dict_input.assert_not_called()
        mock_handle_dataframe_input.assert_called_once()  # default

    # - handle input examples

    def test_handle_dataframe_input_pandas_dataframe(
        self,
        input_example_dataframe_pandas_dataframe,
        input_example_dataframe_pandas_dataframe_empty,
        input_example_dataframe_list,
    ):
        # Act
        json = util._handle_dataframe_input(input_example_dataframe_pandas_dataframe)
        with pytest.raises(ValueError) as e_info:
            util._handle_dataframe_input(input_example_dataframe_pandas_dataframe_empty)

        # Assert
        assert isinstance(json, list)
        assert json == input_example_dataframe_list
        assert "can not be empty" in str(e_info.value)

    def test_handle_dataframe_input_pandas_dataframe_series(
        self,
        input_example_dataframe_pandas_series,
        input_example_dataframe_pandas_series_empty,
        input_example_dataframe_list,
    ):
        # Act
        json = util._handle_dataframe_input(input_example_dataframe_pandas_series)
        with pytest.raises(ValueError) as e_info:
            util._handle_dataframe_input(input_example_dataframe_pandas_series_empty)

        # Assert
        assert isinstance(json, list)
        assert json == input_example_dataframe_list
        assert "can not be empty" in str(e_info.value)

    def test_handle_dataframe_input_list(self, input_example_dataframe_list):
        # Act
        json = util._handle_dataframe_input(input_example_dataframe_list)

        # Assert
        assert isinstance(json, list)
        assert json == input_example_dataframe_list

    def test_handle_dataframe_input_unsupported(self):
        # Act
        with pytest.raises(TypeError) as e_info:
            util._handle_dataframe_input(lambda unsupported: None)

        # Assert
        assert "is not a supported input example type" in str(e_info.value)

    def test_handle_tensor_input(
        self, input_example_numpy, input_example_dataframe_list
    ):
        # Act
        json = util._handle_tensor_input(input_example_numpy)

        # Assert
        assert isinstance(json, list)
        assert json == input_example_dataframe_list

    def test_handle_dict_input(self, input_example_dict):
        # Act
        json = util._handle_dict_input(input_example_dict)

        # Assert
        assert isinstance(json, dict)
        assert json == input_example_dict

    # artifacts

    def test_compress_dir(self, mocker):
        # Arrange
        archive_name = "archive_name"
        path_to_archive = os.path.join("this", "is", "the", "path", "to", "archive")
        archive_out_path = os.path.join(
            "this", "is", "the", "output", "path", "to", "archive"
        )
        full_archive_out_path = os.path.join(archive_out_path, archive_name)
        mock_isdir = mocker.patch("os.path.isdir", return_value=True)
        mock_shutil_make_archive = mocker.patch(
            "shutil.make_archive", return_value="resulting_path"
        )

        # Act
        path = util.compress(archive_out_path, archive_name, path_to_archive)

        # Assert
        assert path == "resulting_path"
        mock_isdir.assert_called_once_with(path_to_archive)
        mock_shutil_make_archive.assert_called_once_with(
            full_archive_out_path, "gztar", path_to_archive
        )

    def test_compress_file(self, mocker):
        # Arrange
        archive_name = "archive_name"
        path_to_archive = os.path.join("path", "to", "archive")
        archive_out_path = os.path.join("output", "path", "to", "archive")
        full_archive_out_path = os.path.join(archive_out_path, archive_name)
        archive_path_dirname = os.path.join("path", "to")
        archive_path_basename = "archive"
        mock_isdir = mocker.patch("os.path.isdir", return_value=False)
        mock_shutil_make_archive = mocker.patch(
            "shutil.make_archive", return_value="resulting_path"
        )

        # Act
        path = util.compress(archive_out_path, archive_name, path_to_archive)

        # Assert
        assert path == "resulting_path"
        mock_isdir.assert_called_once_with(path_to_archive)
        mock_shutil_make_archive.assert_called_once_with(
            full_archive_out_path, "gztar", archive_path_dirname, archive_path_basename
        )

    def test_decompress(self, mocker):
        # Arrange
        archive_file_path = os.path.join("path", "to", "archive", "file")
        extract_dir = False
        mock_shutil_unpack_archive = mocker.patch(
            "shutil.unpack_archive", return_value="resulting_path"
        )

        # Act
        path = util.decompress(archive_file_path, extract_dir)

        # Assert
        assert path == "resulting_path"
        mock_shutil_unpack_archive.assert_called_once_with(
            archive_file_path, extract_dir=extract_dir
        )

    # export models

    def test_validate_metrics(self, model_metrics):
        # Act
        util.validate_metrics(model_metrics)

        # Assert
        # noop

    def test_validate_metrics_unsupported_type(self, model_metrics_wrong_type):
        # Act
        with pytest.raises(TypeError) as e_info:
            util.validate_metrics(model_metrics_wrong_type)

        # Assert
        assert "expected a dict" in str(e_info.value)

    def test_validate_metrics_unsupported_metric_type(
        self, model_metrics_wrong_metric_type
    ):
        # Act
        with pytest.raises(TypeError) as e_info:
            util.validate_metrics(model_metrics_wrong_metric_type)

        # Assert
        assert "expected a string" in str(e_info.value)

    def test_validate_metrics_unsupported_metric_value(
        self, model_metrics_wrong_metric_value
    ):
        # Act
        with pytest.raises(ValueError) as e_info:
            util.validate_metrics(model_metrics_wrong_metric_value)

        # Assert
        assert "is not a number" in str(e_info.value)

    # model serving

    def test_get_predictor_for_model_base(self, mocker, model_base):
        # Arrange
        def pred_base_spec(model_framework, model_server):
            pass

        pred_base = mocker.patch(
            "hsml.predictor.Predictor.__init__", return_value=None, spec=pred_base_spec
        )
        pred_python = mocker.patch("hsml.python.predictor.Predictor.__init__")
        pred_sklearn = mocker.patch("hsml.sklearn.predictor.Predictor.__init__")
        pred_tensorflow = mocker.patch("hsml.tensorflow.predictor.Predictor.__init__")
        pred_torch = mocker.patch("hsml.torch.predictor.Predictor.__init__")
        pred_llm = mocker.patch("hsml.llm.predictor.Predictor.__init__")

        # Act
        predictor = util.get_predictor_for_model(model_base)

        # Assert
        assert isinstance(predictor, BasePredictor)
        pred_base.assert_called_once_with(
            model_framework=MODEL.FRAMEWORK_PYTHON, model_server=MODEL.FRAMEWORK_PYTHON
        )
        pred_python.assert_not_called()
        pred_sklearn.assert_not_called()
        pred_tensorflow.assert_not_called()
        pred_torch.assert_not_called()
        pred_llm.assert_not_called()

    def test_get_predictor_for_model_python(self, mocker, model_python):
        # Arrange
        pred_base = mocker.patch("hsml.predictor.Predictor.__init__")
        pred_python = mocker.patch(
            "hsml.python.predictor.Predictor.__init__", return_value=None
        )
        pred_sklearn = mocker.patch("hsml.sklearn.predictor.Predictor.__init__")
        pred_tensorflow = mocker.patch("hsml.tensorflow.predictor.Predictor.__init__")
        pred_torch = mocker.patch("hsml.torch.predictor.Predictor.__init__")
        pred_llm = mocker.patch("hsml.llm.predictor.Predictor.__init__")

        # Act
        predictor = util.get_predictor_for_model(model_python)

        # Assert
        assert isinstance(predictor, PyPredictor)
        pred_base.assert_not_called()
        pred_python.assert_called_once()
        pred_sklearn.assert_not_called()
        pred_tensorflow.assert_not_called()
        pred_torch.assert_not_called()
        pred_llm.assert_not_called()

    def test_get_predictor_for_model_sklearn(self, mocker, model_sklearn):
        # Arrange
        pred_base = mocker.patch("hsml.predictor.Predictor.__init__")
        pred_python = mocker.patch("hsml.python.predictor.Predictor.__init__")
        pred_sklearn = mocker.patch(
            "hsml.sklearn.predictor.Predictor.__init__", return_value=None
        )
        pred_tensorflow = mocker.patch("hsml.tensorflow.predictor.Predictor.__init__")
        pred_torch = mocker.patch("hsml.torch.predictor.Predictor.__init__")
        pred_llm = mocker.patch("hsml.llm.predictor.Predictor.__init__")

        # Act
        predictor = util.get_predictor_for_model(model_sklearn)

        # Assert
        assert isinstance(predictor, SkLearnPredictor)
        pred_base.assert_not_called()
        pred_python.assert_not_called()
        pred_sklearn.assert_called_once()
        pred_tensorflow.assert_not_called()
        pred_torch.assert_not_called()
        pred_llm.assert_not_called()

    def test_get_predictor_for_model_tensorflow(self, mocker, model_tensorflow):
        # Arrange
        pred_base = mocker.patch("hsml.predictor.Predictor.__init__")
        pred_python = mocker.patch("hsml.python.predictor.Predictor.__init__")
        pred_sklearn = mocker.patch("hsml.sklearn.predictor.Predictor.__init__")
        pred_tensorflow = mocker.patch(
            "hsml.tensorflow.predictor.Predictor.__init__", return_value=None
        )
        pred_torch = mocker.patch("hsml.torch.predictor.Predictor.__init__")
        pred_llm = mocker.patch("hsml.llm.predictor.Predictor.__init__")

        # Act
        predictor = util.get_predictor_for_model(model_tensorflow)

        # Assert
        assert isinstance(predictor, TFPredictor)
        pred_base.assert_not_called()
        pred_python.assert_not_called()
        pred_sklearn.assert_not_called()
        pred_tensorflow.assert_called_once()
        pred_torch.assert_not_called()
        pred_llm.assert_not_called()

    def test_get_predictor_for_model_torch(self, mocker, model_torch):
        # Arrange
        pred_base = mocker.patch("hsml.predictor.Predictor.__init__")
        pred_python = mocker.patch("hsml.python.predictor.Predictor.__init__")
        pred_sklearn = mocker.patch("hsml.sklearn.predictor.Predictor.__init__")
        pred_tensorflow = mocker.patch("hsml.tensorflow.predictor.Predictor.__init__")
        pred_torch = mocker.patch(
            "hsml.torch.predictor.Predictor.__init__", return_value=None
        )
        pred_llm = mocker.patch("hsml.llm.predictor.Predictor.__init__")

        # Act
        predictor = util.get_predictor_for_model(model_torch)

        # Assert
        assert isinstance(predictor, TorchPredictor)
        pred_base.assert_not_called()
        pred_python.assert_not_called()
        pred_sklearn.assert_not_called()
        pred_tensorflow.assert_not_called()
        pred_torch.assert_called_once()
        pred_llm.assert_not_called()

    def test_get_predictor_for_model_llm(self, mocker, model_llm):
        # Arrange
        pred_base = mocker.patch("hsml.predictor.Predictor.__init__")
        pred_python = mocker.patch("hsml.python.predictor.Predictor.__init__")
        pred_sklearn = mocker.patch("hsml.sklearn.predictor.Predictor.__init__")
        pred_tensorflow = mocker.patch("hsml.tensorflow.predictor.Predictor.__init__")
        pred_torch = mocker.patch("hsml.torch.predictor.Predictor.__init__")
        pred_llm = mocker.patch(
            "hsml.llm.predictor.Predictor.__init__", return_value=None
        )

        # Act
        predictor = util.get_predictor_for_model(model_llm)

        # Assert
        assert isinstance(predictor, LLMPredictor)
        pred_base.assert_not_called()
        pred_python.assert_not_called()
        pred_sklearn.assert_not_called()
        pred_tensorflow.assert_not_called()
        pred_torch.assert_not_called()
        pred_llm.assert_called_once()

    def test_get_predictor_for_model_non_base(self, mocker):
        # Arrange
        pred_base = mocker.patch("hsml.predictor.Predictor.__init__")
        pred_python = mocker.patch("hsml.python.predictor.Predictor.__init__")
        pred_sklearn = mocker.patch("hsml.sklearn.predictor.Predictor.__init__")
        pred_tensorflow = mocker.patch("hsml.tensorflow.predictor.Predictor.__init__")
        pred_torch = mocker.patch("hsml.torch.predictor.Predictor.__init__")
        pred_llm = mocker.patch("hsml.llm.predictor.Predictor.__init__")

        class NonBaseModel:
            pass

        # Act
        with pytest.raises(ValueError) as e_info:
            util.get_predictor_for_model(NonBaseModel())

        assert "an instance of {} class is expected".format(BaseModel) in str(
            e_info.value
        )
        pred_base.assert_not_called()
        pred_python.assert_not_called()
        pred_sklearn.assert_not_called()
        pred_tensorflow.assert_not_called()
        pred_torch.assert_not_called()
        pred_llm.assert_not_called()

    def test_get_hostname_replaced_url(self, mocker):
        # Arrange
        sub_path = "this/is/a/sub_path"
        base_url = "/hopsworks/api/base/"
        urlparse_href_arg = ParseResult(
            scheme="",
            netloc="",
            path=base_url + sub_path,
            params="",
            query="",
            fragment="",
        )
        geturl_return = "final_url"
        mock_url_parsed = mocker.MagicMock()
        mock_url_parsed.geturl = mocker.MagicMock(return_value=geturl_return)
        mock_client = mocker.MagicMock()
        mock_client._base_url = base_url + "url"
        mock_client.replace_public_host = mocker.MagicMock(return_value=mock_url_parsed)
        mocker.patch("hsml.client.get_instance", return_value=mock_client)

        # Act
        url = util.get_hostname_replaced_url(sub_path)

        # Assert
        mock_client.replace_public_host.assert_called_once_with(urlparse_href_arg)
        mock_url_parsed.geturl.assert_called_once()
        assert url == geturl_return

    # general

    def test_get_members(self):
        # Arrange
        class TEST:
            TEST_1 = 1
            TEST_2 = "two"
            TEST_3 = "3"

        # Act
        members = list(util.get_members(TEST))

        # Assert
        assert members == [1, "two", "3"]

    def test_get_members_with_prefix(self):
        # Arrange
        class TEST:
            TEST_1 = 1
            TEST_2 = "two"
            RES_3 = "3"
            NONE = None

        # Act
        members = list(util.get_members(TEST, prefix="TEST"))

        # Assert
        assert members == [1, "two"]

    # json

    def test_extract_field_from_json(self, mocker):
        # Arrange
        json = {"a": "1", "b": "2"}
        get_obj_from_json = mocker.patch("hsml.util.get_obj_from_json")

        # Act
        b = util.extract_field_from_json(json, "b")

        # Assert
        assert b == "2"
        assert get_obj_from_json.call_count == 0

    def test_extract_field_from_json_fields(self, mocker):
        # Arrange
        json = {"a": "1", "b": "2"}
        get_obj_from_json = mocker.patch("hsml.util.get_obj_from_json")

        # Act
        b = util.extract_field_from_json(json, ["B", "b"])  # alternative fields

        # Assert
        assert b == "2"
        assert get_obj_from_json.call_count == 0

    def test_extract_field_from_json_as_instance_of_str(self, mocker):
        # Arrange
        json = {"a": "1", "b": "2"}
        get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value="2"
        )

        # Act
        b = util.extract_field_from_json(json, "b", as_instance_of=str)

        # Assert
        assert b == "2"
        get_obj_from_json.assert_called_once_with(obj="2", cls=str)

    def test_extract_field_from_json_as_instance_of_list_str(self, mocker):
        # Arrange
        json = {"a": "1", "b": ["2", "2", "2"]}
        get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value="2"
        )

        # Act
        b = util.extract_field_from_json(json, "b", as_instance_of=str)

        # Assert
        assert b == ["2", "2", "2"]
        assert get_obj_from_json.call_count == 3
        assert get_obj_from_json.call_args[1]["obj"] == "2"
        assert get_obj_from_json.call_args[1]["cls"] is str

    def test_get_obj_from_json_cls(self, mocker):
        # Arrange
        class Test:
            def __init__(self):
                self.a = "1"

        # Act
        obj = util.get_obj_from_json(Test(), Test)

        # Assert
        assert isinstance(obj, Test)
        assert obj.a == "1"

    def test_get_obj_from_json_dict(self, mocker):
        # Arrange
        class Test:
            def __init__(self, a):
                self.a = a

            @classmethod
            def from_json(cls, json):
                return cls(**json)

        # Act
        obj = util.get_obj_from_json({"a": "1"}, Test)

        # Assert
        assert isinstance(obj, Test)
        assert obj.a == "1"

    def test_get_obj_from_json_dict_default(self, mocker):
        # Arrange
        class Test:
            def __init__(self, a="11"):
                self.a = "11"

            @classmethod
            def from_json(cls, json):
                return cls(**json)

        # Act
        obj = util.get_obj_from_json({}, Test)

        # Assert
        assert isinstance(obj, Test)
        assert obj.a == "11"

    def test_get_obj_from_json_unsupported(self, mocker):
        # Arrange
        class Test:
            pass

        # Act
        with pytest.raises(ValueError) as e_info:
            util.get_obj_from_json("UNSUPPORTED", Test)

        # Assert
        assert "cannot be converted to class" in str(e_info.value)

    def test_get_hudi_datestr_from_timestamp(self):
        dt = hsfs.util.get_hudi_datestr_from_timestamp(1640995200000)
        assert dt == "20220101000000000"

    def test_convert_event_time_to_timestamp_timestamp(self):
        dt = hsfs.util.convert_event_time_to_timestamp(1640995200)
        assert dt == 1640995200000

    def test_convert_event_time_to_timestamp_datetime(self):
        dt = hsfs.util.convert_event_time_to_timestamp(datetime(2022, 1, 1, 0, 0, 0))
        assert dt == 1640995200000

    def test_convert_event_time_to_timestamp_datetime_tz(self):
        dt = hsfs.util.convert_event_time_to_timestamp(
            pytz.timezone("US/Pacific").localize(datetime(2021, 12, 31, 16, 0, 0))
        )
        assert dt == 1640995200000

    def test_convert_event_time_to_timestamp_date(self):
        dt = hsfs.util.convert_event_time_to_timestamp(date(2022, 1, 1))
        assert dt == 1640995200000

    def test_convert_event_time_to_timestamp_string(self):
        dt = hsfs.util.convert_event_time_to_timestamp("2022-01-01 00:00:00")
        assert dt == 1640995200000

    def test_convert_iso_event_time_to_timestamp_string(self):
        dt = hsfs.util.convert_event_time_to_timestamp("2022-01-01T00:00:00.000000Z")
        assert dt == 1640995200000

    def test_convert_event_time_to_timestamp_yyyy_mm_dd(self):
        timestamp = hsfs.util.get_timestamp_from_date_string("2022-01-01")
        assert timestamp == 1640995200000

    def test_convert_event_time_to_timestamp_yyyy_mm_dd_hh(self):
        timestamp = hsfs.util.get_timestamp_from_date_string("2022-01-01 00")
        assert timestamp == 1640995200000

    def test_convert_event_time_to_timestamp_yyyy_mm_dd_hh_mm(self):
        timestamp = hsfs.util.get_timestamp_from_date_string("2022-01-01 00:00")
        assert timestamp == 1640995200000

    def test_convert_event_time_to_timestamp_yyyy_mm_dd_hh_mm_ss(self):
        timestamp = hsfs.util.get_timestamp_from_date_string("2022-01-01 00:00:00")
        assert timestamp == 1640995200000

    def test_convert_event_time_to_timestamp_yyyy_mm_dd_hh_mm_ss_f(self):
        timestamp = hsfs.util.get_timestamp_from_date_string("2022-01-01 00:00:00.000")
        assert timestamp == 1640995200000

    def test_convert_event_time_to_timestamp_yyyy_mm_dd_hh_mm_ss_error(self):
        with pytest.raises(ValueError):
            hsfs.util.get_timestamp_from_date_string("2022-13-01 00:00:00")

    def test_convert_event_time_to_timestamp_yyyy_mm_dd_hh_mm_ss_error2(self):
        with pytest.raises(ValueError):
            hsfs.util.get_timestamp_from_date_string("202-13-01 00:00:00")

    def test_convert_event_time_to_timestamp_yyyy_mm_dd_hh_mm_ss_error3(self):
        with pytest.raises(ValueError):
            hsfs.util.get_timestamp_from_date_string("00:00:00 2022-01-01")

    def test_convert_hudi_commit_time_to_timestamp(self):
        timestamp = hsfs.util.get_timestamp_from_date_string("20221118095233099")
        assert timestamp == 1668765153099

    def test_get_dataset_type_HIVEDB(self):
        db_type = hsfs.util.get_dataset_type(
            "/apps/hive/warehouse/temp_featurestore.db/storage_connector_resources/kafka__tstore.jks"
        )
        assert db_type == "HIVEDB"

    def test_get_dataset_type_HIVEDB_with_dfs(self):
        db_type = hsfs.util.get_dataset_type(
            "hdfs:///apps/hive/warehouse/temp_featurestore.db/storage_connector_resources/kafka__tstore.jks"
        )
        assert db_type == "HIVEDB"

    def test_get_dataset_type_DATASET(self):
        db_type = hsfs.util.get_dataset_type(
            "/Projects/temp/Resources/kafka__tstore.jks"
        )
        assert db_type == "DATASET"

    def test_get_dataset_type_DATASET_with_dfs(self):
        db_type = hsfs.util.get_dataset_type(
            "hdfs:///Projects/temp/Resources/kafka__tstore.jks"
        )
        assert db_type == "DATASET"

    def test_get_job_url(self, mocker):
        # Arrange
        mock_client_get_instance = mocker.patch("hopsworks_common.client.get_instance")

        # Act
        hsfs.util.get_job_url(href="1/2/3/4/5/6/7/8")

        # Assert
        assert (
            mock_client_get_instance.return_value.replace_public_host.call_args[0][
                0
            ].path
            == "p/5/jobs/named/7/executions"
        )

    def test_get_feature_group_url(self, mocker):
        # Arrange
        feature_store_id = 99
        feature_group_id = 10
        mock_client_get_instance = mocker.patch("hopsworks_common.client.get_instance")
        mock_util_get_hostname_replaced_url = mocker.patch(
            "hopsworks_common.util.get_hostname_replaced_url"
        )
        mock_client_get_instance.return_value._project_id = 50

        # Act
        hsfs.util.get_feature_group_url(
            feature_group_id=feature_group_id, feature_store_id=feature_store_id
        )

        # Assert
        assert mock_util_get_hostname_replaced_url.call_count == 1
        assert (
            mock_util_get_hostname_replaced_url.call_args[0][0] == "/p/50/fs/99/fg/10"
        )

    def test_valid_embedding_type(self):
        embedding_index = EmbeddingIndex(
            features=[
                EmbeddingFeature("feature1", 3),
                EmbeddingFeature("feature2", 3),
                EmbeddingFeature("feature3", 3),
                EmbeddingFeature("feature4", 3),
            ]
        )
        # Define a schema with valid feature types
        schema = [
            Feature(name="feature1", type="array<int>"),
            Feature(name="feature2", type="array<bigint>"),
            Feature(name="feature3", type="array<float>"),
            Feature(name="feature4", type="array<double>"),
        ]
        # Call the method and expect no exceptions
        hsfs.util.validate_embedding_feature_type(embedding_index, schema)

    def test_invalid_embedding_type(self):
        embedding_index = EmbeddingIndex(
            features=[
                EmbeddingFeature("feature1", 3),
                EmbeddingFeature("feature2", 3),
            ]
        )
        # Define a schema with an invalid feature type
        schema = [
            Feature(name="feature1", type="array<int>"),
            Feature(name="feature2", type="array<string>"),  # Invalid type
        ]
        # Call the method and expect a FeatureStoreException
        with pytest.raises(FeatureStoreException):
            hsfs.util.validate_embedding_feature_type(embedding_index, schema)

    def test_missing_embedding_index(self):
        # Define a schema without an embedding index
        schema = [
            Feature(name="feature1", type="array<int>"),
            Feature(name="feature2", type="array<bigint>"),
        ]
        # Call the method with an empty feature_group (no embedding index)
        hsfs.util.validate_embedding_feature_type(None, schema)
        # No exception should be raised

    def test_empty_schema(self):
        embedding_index = EmbeddingIndex(
            features=[
                EmbeddingFeature("feature1", 3),
                EmbeddingFeature("feature2", 3),
            ]
        )
        # Define an empty schema
        schema = []
        # Call the method with an empty schema
        hsfs.util.validate_embedding_feature_type(embedding_index, schema)
        # No exception should be raised

    @pytest.mark.skipif(
        not HAS_SQLALCHEMY or not HAS_AIOMYSQL,
        reason="SQLAlchemy or aiomysql is not installed",
    )
    def test_create_async_engine(self, mocker):
        # Test when get_running_loop() raises a RuntimeError
        with patch("asyncio.get_running_loop", side_effect=RuntimeError):
            # mock storage connector
            online_connector = patch.object(hsfs.util, "get_online_connector")
            with pytest.raises(
                RuntimeError,
                match="Event loop is not running. Please invoke this co-routine from a running loop or provide an event loop.",
            ):
                asyncio.run(util_sql.create_async_engine(online_connector, True, 1))



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from typing import Optional

import humps
from hsml import util


class PredictorStateCondition:
    """Condition of a predictor state."""

    def __init__(
        self,
        type: str,
        status: Optional[bool] = None,
        reason: Optional[str] = None,
        **kwargs,
    ):
        self._type = type
        self._status = status
        self._reason = reason

    def describe(self):
        util.pretty_print(self)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls.from_json(json_decamelized)

    @classmethod
    def from_json(cls, json_decamelized):
        return PredictorStateCondition(**cls.extract_fields_from_json(json_decamelized))

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        kwargs = {}
        kwargs["type"] = json_decamelized.pop("type")  # required
        kwargs["status"] = util.extract_field_from_json(json_decamelized, "status")
        kwargs["reason"] = util.extract_field_from_json(json_decamelized, "reason")
        return kwargs

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**self.extract_fields_from_json(json_decamelized))
        return self

    def json(self):
        return json.dumps(self, cls=util.MLEncoder)

    def to_dict(self):
        return {
            "condition": {
                "type": self._type,
                "status": self._status,
                "reason": self._reason,
            }
        }

    @property
    def type(self):
        """Condition type of the predictor state."""
        return self._type

    @property
    def status(self):
        """Condition status of the predictor state."""
        return self._status

    @property
    def reason(self):
        """Condition reason of the predictor state."""
        return self._reason

    def __repr__(self):
        return f"PredictorStateCondition(type: {self.type.capitalize()!r}, status: {self.status!r})"



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hsml import deployable_component, inference_batcher


class TestDeployableComponent:
    # from response json

    def test_from_response_json(self, mocker):
        # Arrange
        json = {"test": "test"}
        mock_from_json = mocker.patch(
            "hsml.deployable_component.DeployableComponent.from_json",
            return_value="from_json_result",
        )

        # Act
        result = deployable_component.DeployableComponent.from_response_json(json)

        # Assert
        assert result == "from_json_result"
        mock_from_json.assert_called_once_with(json)

    # constructor

    def test_constructor_default(self, mocker):
        # Arrange
        mock_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=None
        )
        mock_ib_init = mocker.patch(
            "hsml.inference_batcher.InferenceBatcher.__init__", return_value=None
        )

        class DeployableComponentChild(deployable_component.DeployableComponent):
            def from_json():
                pass

            def update_from_response_json():
                pass

            def to_dict():
                pass

        # Act
        dc = DeployableComponentChild()

        # Assert
        assert dc.script_file is None
        assert dc.resources is None
        mock_get_obj_from_json.assert_called_once_with(
            None, inference_batcher.InferenceBatcher
        )
        mock_ib_init.assert_called_once()

    def test_constructor_with_params(self, mocker):
        # Arrange
        script_file = "script_file"
        resources = {}
        inf_batcher = inference_batcher.InferenceBatcher()
        mock_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=inf_batcher
        )
        mock_ib_init = mocker.patch(
            "hsml.inference_batcher.InferenceBatcher.__init__", return_value=None
        )

        class DeployableComponentChild(deployable_component.DeployableComponent):
            def from_json():
                pass

            def update_from_response_json():
                pass

            def to_dict():
                pass

        # Act
        dc = DeployableComponentChild(
            script_file=script_file,
            resources=resources,
            inference_batcher=inf_batcher,
        )

        # Assert
        assert dc.script_file == script_file
        assert dc.resources == resources
        mock_get_obj_from_json.assert_called_once_with(
            inf_batcher, inference_batcher.InferenceBatcher
        )
        assert dc.inference_batcher == inf_batcher
        mock_ib_init.assert_not_called()



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from hopsworks_common.client.exceptions import FeatureStoreException
from hsfs import feature_group_commit, util
from hsfs.core import feature_group_api


try:
    from delta.tables import DeltaTable
except ImportError:
    pass


class DeltaEngine:
    DELTA_SPARK_FORMAT = "delta"
    DELTA_QUERY_TIME_TRAVEL_AS_OF_INSTANT = "timestampAsOf"

    def __init__(
        self,
        feature_store_id,
        feature_store_name,
        feature_group,
        spark_session,
        spark_context,
    ):
        self._feature_group = feature_group
        self._spark_context = spark_context
        self._spark_session = spark_session
        self._feature_store_id = feature_store_id
        self._feature_store_name = feature_store_name

        self._feature_group_api = feature_group_api.FeatureGroupApi()

    def save_delta_fg(self, dataset, write_options, validation_id=None):
        fg_commit = self._write_delta_dataset(dataset, write_options)
        fg_commit.validation_id = validation_id
        return self._feature_group_api.commit(self._feature_group, fg_commit)

    def register_temporary_table(self, delta_fg_alias, read_options):
        location = self._feature_group.prepare_spark_location()

        delta_options = self._setup_delta_read_opts(delta_fg_alias, read_options)
        self._spark_session.read.format(self.DELTA_SPARK_FORMAT).options(
            **delta_options
        ).load(location).createOrReplaceTempView(
            delta_fg_alias.alias
        )

    def _setup_delta_read_opts(self, delta_fg_alias, read_options):
        delta_options = {}
        if delta_fg_alias.left_feature_group_end_timestamp is None and (
            delta_fg_alias.left_feature_group_start_timestamp is None
            or delta_fg_alias.left_feature_group_start_timestamp == 0
        ):
            # snapshot query latest state
            delta_options = {}
        elif (
            delta_fg_alias.left_feature_group_end_timestamp is not None
            and delta_fg_alias.left_feature_group_start_timestamp is None
        ):
            # snapshot query with end time
            _delta_commit_end_time = util.get_delta_datestr_from_timestamp(
                delta_fg_alias.left_feature_group_end_timestamp
            )
            delta_options = {
                self.DELTA_QUERY_TIME_TRAVEL_AS_OF_INSTANT: _delta_commit_end_time,
            }

        if read_options:
            delta_options.update(read_options)

        return delta_options

    def delete_record(self, delete_df):
        location = self._feature_group.prepare_spark_location()

        if not DeltaTable.isDeltaTable(
            self._spark_session, location
        ):
            raise FeatureStoreException(
                f"This is no data available in Feature group {self._feature_group.name}, or it not DELTA enabled "
            )
        else:
            fg_source_table = DeltaTable.forPath(
                self._spark_session, location
            )

            source_alias = (
                f"{self._feature_group.name}_{self._feature_group.version}_source"
            )
            updates_alias = (
                f"{self._feature_group.name}_{self._feature_group.version}_updates"
            )
            merge_query_str = self._generate_merge_query(source_alias, updates_alias)

            fg_source_table.alias(source_alias).merge(
                delete_df.alias(updates_alias), merge_query_str
            ).whenMatchedDelete().execute()

        fg_commit = self._get_last_commit_metadata(
            self._spark_session, location
        )
        return self._feature_group_api.commit(self._feature_group, fg_commit)

    def _write_delta_dataset(self, dataset, write_options):
        location = self._feature_group.prepare_spark_location()

        if write_options is None:
            write_options = {}

        if not DeltaTable.isDeltaTable(
            self._spark_session, location
        ):
            (
                dataset.write.format(DeltaEngine.DELTA_SPARK_FORMAT)
                .options(**write_options)
                .partitionBy(
                    self._feature_group.partition_key
                    if self._feature_group.partition_key
                    else []
                )
                .mode("append")
                .save(location)
            )
        else:
            fg_source_table = DeltaTable.forPath(
                self._spark_session, location
            )

            source_alias = (
                f"{self._feature_group.name}_{self._feature_group.version}_source"
            )
            updates_alias = (
                f"{self._feature_group.name}_{self._feature_group.version}_updates"
            )
            merge_query_str = self._generate_merge_query(source_alias, updates_alias)

            fg_source_table.alias(source_alias).merge(
                dataset.alias(updates_alias), merge_query_str
            ).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()

        return self._get_last_commit_metadata(
            self._spark_session, location
        )

    def vacuum(self, retention_hours: int):
        location = self._feature_group.prepare_spark_location()
        retention = f"RETAIN {retention_hours} HOURS" if retention_hours is not None else ""
        self._spark_session.sql(f"VACUUM '{location}' {retention}")

    def _generate_merge_query(self, source_alias, updates_alias):
        merge_query_list = []
        primary_key = self._feature_group.primary_key

        # add event time to primary key for upserts
        if self._feature_group.event_time is not None:
            primary_key.append(self._feature_group.event_time)

        # add partition key for upserts
        if self._feature_group.partition_key:
            primary_key = primary_key + self._feature_group.partition_key

        for pk in primary_key:
            merge_query_list.append(f"{source_alias}.{pk} == {updates_alias}.{pk}")
        megrge_query_str = " AND ".join(merge_query_list)
        return megrge_query_str

    @staticmethod
    def _get_last_commit_metadata(spark_context, base_path):
        fg_source_table = DeltaTable.forPath(spark_context, base_path)

        # Get info about the latest commit
        last_commit = fg_source_table.history(1).first().asDict()
        version = last_commit["version"]
        commit_timestamp = util.convert_event_time_to_timestamp(
            last_commit["timestamp"]
        )
        commit_date_string = util.get_hudi_datestr_from_timestamp(commit_timestamp)
        operation_metrics = last_commit["operationMetrics"]

        # Get info about the oldest remaining commit
        oldest_commit = fg_source_table.history().orderBy("version").first().asDict()
        oldest_commit_timestamp = util.convert_event_time_to_timestamp(
            oldest_commit["timestamp"]
        )

        if version == 0:
            fg_commit = feature_group_commit.FeatureGroupCommit(
                commitid=None,
                commit_date_string=commit_date_string,
                commit_time=commit_timestamp,
                rows_inserted=operation_metrics["numOutputRows"],
                rows_updated=0,
                rows_deleted=0,
                last_active_commit_time=oldest_commit_timestamp,
            )
        else:
            fg_commit = feature_group_commit.FeatureGroupCommit(
                commitid=None,
                commit_date_string=commit_date_string,
                commit_time=commit_timestamp,
                rows_inserted=operation_metrics["numTargetRowsInserted"],
                rows_updated=operation_metrics["numTargetRowsUpdated"],
                rows_deleted=operation_metrics["numTargetRowsDeleted"],
                last_active_commit_time=oldest_commit_timestamp,
            )

        return fg_commit



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Any, Dict, Union

import humps
from hsfs import feature_group


class ExternalFeatureGroupAlias:
    def __init__(
        self, on_demand_feature_group: Dict[str, Any], alias: str, **kwargs
    ) -> None:
        self._on_demand_feature_group: Union[
            feature_group.ExternalFeatureGroup, "feature_group.SpineGroup"
        ]
        if not on_demand_feature_group["spine"]:
            self._on_demand_feature_group = (
                feature_group.ExternalFeatureGroup.from_response_json(
                    on_demand_feature_group
                )
            )
        else:
            self._on_demand_feature_group = feature_group.SpineGroup.from_response_json(
                on_demand_feature_group
            )
        self._alias = alias

    @classmethod
    def from_response_json(cls, json_dict: Dict[str, Any]) -> ExternalFeatureGroupAlias:
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    @property
    def on_demand_feature_group(
        self,
    ) -> Union[feature_group.ExternalFeatureGroup, "feature_group.SpineGroup"]:
        return self._on_demand_feature_group

    @property
    def alias(self) -> str:
        return self._alias



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

from typing import Union

import hopsworks_common.client as _main
from hopsworks_common.client.istio import external, hopsworks


_client: Union[hopsworks.Client, external.Client, None] = None


def init(host, port, project=None, api_key_value=None):
    global _client

    if _client:
        return
    if not _main._is_external():
        _client = hopsworks.Client(host, port)
    else:
        _client = external.Client(host, port, project, api_key_value)


def get_instance() -> Union[hopsworks.Client, external.Client, None]:
    global _client
    return _client



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import feature_group_commit


class TestFeatureGroupCommit:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["feature_group_commit"]["get_list"]["response"]

        # Act
        fg_commit = feature_group_commit.FeatureGroupCommit.from_response_json(json)[0]

        # Assert
        assert fg_commit.commitid == 11
        assert fg_commit.commit_date_string == "test_commit_date_string"
        assert fg_commit.commit_time == "test_commit_time"
        assert fg_commit.rows_inserted == 1
        assert fg_commit.rows_updated == 2
        assert fg_commit.rows_deleted == 3
        assert fg_commit.validation_id == 77

    def test_from_response_json_list(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["feature_group_commit"]["get_list"]["response"]

        # Act
        fg_commit_list = feature_group_commit.FeatureGroupCommit.from_response_json(
            json
        )

        # Assert
        assert len(fg_commit_list) == 1
        fg_commit = fg_commit_list[0]
        assert fg_commit.commitid == 11
        assert fg_commit.commit_date_string == "test_commit_date_string"
        assert fg_commit.commit_time == "test_commit_time"
        assert fg_commit.rows_inserted == 1
        assert fg_commit.rows_updated == 2
        assert fg_commit.rows_deleted == 3
        assert fg_commit.validation_id == 77

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["feature_group_commit"]["get_basic_info"]["response"]

        # Act
        fg_commit = feature_group_commit.FeatureGroupCommit.from_response_json(json)

        # Assert
        assert len(fg_commit) == 0



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# This implementation has been borrowed from the kserve/kserve repository
# https://github.com/kserve/kserve/blob/release-0.11/python/kserve/kserve/errors.py


class InvalidInput(ValueError):
    """
    Exception class indicating invalid input arguments.
    HTTP Servers should return HTTP_400 (Bad Request).
    """

    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.reason



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import copy

import humps
import pytest
from hsml import kafka_topic
from hsml.constants import KAFKA_TOPIC


class TestKafkaTopic:
    # from response json

    def test_from_response_json_with_name_only(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_existing_with_name_only"][
            "response"
        ]["kafka_topic_dto"]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_kt_from_json = mocker.patch("hsml.kafka_topic.KafkaTopic.from_json")

        # Act
        _ = kafka_topic.KafkaTopic.from_response_json(json_camelized)

        # Assert
        mock_kt_from_json.assert_called_once_with(json)

    def test_from_response_json_with_name_and_config(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_existing_with_name_and_config"][
            "response"
        ]["kafka_topic_dto"]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_kt_from_json = mocker.patch("hsml.kafka_topic.KafkaTopic.from_json")

        # Act
        _ = kafka_topic.KafkaTopic.from_response_json(json_camelized)

        # Assert
        mock_kt_from_json.assert_called_once_with(json)

    # from json

    def test_from_json_with_name_only(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_existing_with_name_only"][
            "response"
        ]["kafka_topic_dto"]
        mock_kt_extract_fields = mocker.patch(
            "hsml.kafka_topic.KafkaTopic.extract_fields_from_json", return_value=json
        )
        mock_kt_init = mocker.patch(
            "hsml.kafka_topic.KafkaTopic.__init__", return_value=None
        )

        # Act
        _ = kafka_topic.KafkaTopic.from_response_json(json)

        # Assert
        mock_kt_extract_fields.assert_called_once_with(json)
        mock_kt_init.assert_called_once_with(**json)

    def test_from_json_with_name_and_config(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_existing_with_name_and_config"][
            "response"
        ]["kafka_topic_dto"]
        mock_kt_extract_fields = mocker.patch(
            "hsml.kafka_topic.KafkaTopic.extract_fields_from_json", return_value=json
        )
        mock_kt_init = mocker.patch(
            "hsml.kafka_topic.KafkaTopic.__init__", return_value=None
        )

        # Act
        _ = kafka_topic.KafkaTopic.from_response_json(json)

        # Assert
        mock_kt_extract_fields.assert_called_once_with(json)
        mock_kt_init.assert_called_once_with(**json)

    # constructor

    def test_constructor_existing_with_name_only(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_existing_with_name_only"][
            "response"
        ]["kafka_topic_dto"]
        mock_kt_validate_topic_config = mocker.patch(
            "hsml.kafka_topic.KafkaTopic._validate_topic_config",
            return_value=(KAFKA_TOPIC.NUM_REPLICAS, KAFKA_TOPIC.NUM_PARTITIONS),
        )

        # Act
        kt = kafka_topic.KafkaTopic(**json)

        # Assert
        assert isinstance(kt, kafka_topic.KafkaTopic)
        assert kt.name == json["name"]
        assert kt.num_replicas == KAFKA_TOPIC.NUM_REPLICAS
        assert kt.num_partitions == KAFKA_TOPIC.NUM_PARTITIONS
        mock_kt_validate_topic_config.assert_called_once_with(json["name"], None, None)

    def test_constructor_existing_with_name_and_config(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_existing_with_name_and_config"][
            "response"
        ]["kafka_topic_dto"]
        mock_kt_validate_topic_config = mocker.patch(
            "hsml.kafka_topic.KafkaTopic._validate_topic_config",
            return_value=(json["num_replicas"], json["num_partitions"]),
        )

        # Act
        kt = kafka_topic.KafkaTopic(**json)

        # Assert
        assert isinstance(kt, kafka_topic.KafkaTopic)
        assert kt.name == json["name"]
        assert kt.num_replicas == json["num_replicas"]
        assert kt.num_partitions == json["num_partitions"]
        mock_kt_validate_topic_config.assert_called_once_with(
            json["name"], json["num_replicas"], json["num_partitions"]
        )

    # validate topic config

    def test_validate_topic_config_existing_with_name_only(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_existing_with_name_only"][
            "response"
        ]["kafka_topic_dto"]

        # Act
        num_repl, num_part = kafka_topic.KafkaTopic._validate_topic_config(
            json["name"], None, None
        )

        # Assert
        assert num_repl is None
        assert num_part is None

    def test_validate_topic_config_existing_with_name_and_config(
        self, backend_fixtures
    ):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_existing_with_name_and_config"][
            "response"
        ]["kafka_topic_dto"]

        # Act
        with pytest.raises(ValueError) as e_info:
            num_repl, num_part = kafka_topic.KafkaTopic._validate_topic_config(
                json["name"], json["num_replicas"], json["num_partitions"]
            )

        # Assert
        assert "Number of replicas or partitions cannot be changed" in str(e_info.value)

    def test_validate_topic_config_none(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_none"]["response"][
            "kafka_topic_dto"
        ]

        # Act
        num_repl, num_part = kafka_topic.KafkaTopic._validate_topic_config(
            json["name"], None, None
        )

        # Assert
        assert num_repl is None
        assert num_part is None

    def test_validate_topic_config_none_with_config(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_none_with_config"]["response"][
            "kafka_topic_dto"
        ]

        # Act
        num_repl, num_part = kafka_topic.KafkaTopic._validate_topic_config(
            json["name"], json["num_replicas"], json["num_partitions"]
        )

        # Assert
        assert num_repl is None
        assert num_part is None

    def test_validate_topic_config_none_value(self):
        # Act
        num_repl, num_part = kafka_topic.KafkaTopic._validate_topic_config(
            None, None, None
        )

        # Assert
        assert num_repl is None
        assert num_part is None

    def test_validate_topic_config_none_value_with_config(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_none_with_config"]["response"][
            "kafka_topic_dto"
        ]

        # Act
        num_repl, num_part = kafka_topic.KafkaTopic._validate_topic_config(
            None, json["num_replicas"], json["num_partitions"]
        )

        # Assert
        assert num_repl is None
        assert num_part is None

    def test_validate_topic_config_create(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_create_with_name_only"]["response"][
            "kafka_topic_dto"
        ]

        # Act
        num_repl, num_part = kafka_topic.KafkaTopic._validate_topic_config(
            json["name"], None, None
        )

        # Assert
        assert num_repl == KAFKA_TOPIC.NUM_REPLICAS
        assert num_part == KAFKA_TOPIC.NUM_PARTITIONS

    def test_validate_topic_config_create_with_config(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_create_with_name_and_config"][
            "response"
        ]["kafka_topic_dto"]

        # Act
        num_repl, num_part = kafka_topic.KafkaTopic._validate_topic_config(
            json["name"], json["num_replicas"], json["num_partitions"]
        )

        # Assert
        assert num_repl == json["num_replicas"]
        assert num_part == json["num_partitions"]

    # extract fields from json

    def test_extract_fields_from_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"]["get_existing_with_name_and_config"][
            "response"
        ]["kafka_topic_dto"]
        json_copy = copy.deepcopy(json)

        # Act
        kwargs = kafka_topic.KafkaTopic.extract_fields_from_json(json_copy)

        # Assert
        assert kwargs["name"] == json["name"]
        assert kwargs["num_replicas"] == json["num_replicas"]
        assert kwargs["num_partitions"] == json["num_partitions"]

    def test_extract_fields_from_json_alternative(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["kafka_topic"][
            "get_existing_with_name_and_config_alternative"
        ]["response"]["kafka_topic_dto"]
        json_copy = copy.deepcopy(json)

        # Act
        kwargs = kafka_topic.KafkaTopic.extract_fields_from_json(json_copy)

        # Assert
        assert kwargs["name"] == json["name"]
        assert kwargs["num_replicas"] == json["num_of_replicas"]
        assert kwargs["num_partitions"] == json["num_of_partitions"]
